"use strict";
Object.defineProperty(exports, "__esModule", {
	value: true
});

var _sdpHelper = require("./sdp-helper");

/*
	The MIT License (MIT)

	Copyright (c) 2016 Meetecho

	Permission is hereby granted, free of charge, to any person obtaining
	a copy of this software and associated documentation files (the "Software"),
	to deal in the Software without restriction, including without limitation
	the rights to use, copy, modify, merge, publish, distribute, sublicense,
	and/or sell copies of the Software, and to permit persons to whom the
	Software is furnished to do so, subject to the following conditions:

	The above copyright notice and this permission notice shall be included
	in all copies or substantial portions of the Software.

	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
	OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
	FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
	THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
	OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
	ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
	OTHER DEALINGS IN THE SOFTWARE.
 */

// List of sessions
Janus.sessions = {};

Janus.isExtensionEnabled = function () {
	if (navigator.mediaDevices && navigator.mediaDevices.getDisplayMedia) {
		// No need for the extension, getDisplayMedia is supported
		return true;
	}
	if (window.navigator.userAgent.match('Chrome')) {
		var chromever = parseInt(window.navigator.userAgent.match(/Chrome\/(.*) /)[1], 10);
		var maxver = 33;
		if (window.navigator.userAgent.match('Linux')) maxver = 35; // "known" crash in chrome 34 and 35 on linux
		if (chromever >= 26 && chromever <= maxver) {
			// Older versions of Chrome don't support this extension-based approach, so lie
			return true;
		}
		return Janus.extension.isInstalled();
	} else {
		// Firefox and others, no need for the extension (but this doesn't mean it will work)
		return true;
	}
};

var defaultExtension = {
	// Screensharing Chrome Extension ID
	extensionId: 'hapfgfdkleiggjjpfpenajgdnfckjpaj',
	isInstalled: function isInstalled() {
		return document.querySelector('#janus-extension-installed') !== null;
	},
	getScreen: function getScreen(callback) {
		var pending = window.setTimeout(function () {
			var error = new Error('NavigatorUserMediaError');
			error.name = 'The required Chrome extension is not installed: click <a href="#">here</a> to install it. (NOTE: this will need you to refresh the page)';
			return callback(error);
		}, 1000);
		this.cache[pending] = callback;
		window.postMessage({ type: 'janusGetScreen', id: pending }, '*');
	},
	init: function init() {
		var cache = {};
		this.cache = cache;
		// Wait for events from the Chrome Extension
		window.addEventListener('message', function (event) {
			if (event.origin != window.location.origin) return;
			if (event.data.type == 'janusGotScreen' && cache[event.data.id]) {
				var callback = cache[event.data.id];
				delete cache[event.data.id];
				if (event.data.sourceId === '') {
					// user canceled
					var error = new Error('NavigatorUserMediaError');
					error.name = 'You cancelled the request for permission, giving up...';
					callback(error);
				} else {
					callback(null, event.data.sourceId);
				}
			} else if (event.data.type == 'janusGetScreenPending') {
				console.log('clearing ', event.data.id);
				window.clearTimeout(event.data.id);
			}
		});
	}
};

Janus.useDefaultDependencies = function (deps) {
	var f = deps && deps.fetch || fetch;
	var p = deps && deps.Promise || Promise;
	var socketCls = deps && deps.WebSocket || WebSocket;

	return {
		newWebSocket: function newWebSocket(server, proto) {
			return new socketCls(server, proto);
		},
		extension: deps && deps.extension || defaultExtension,
		isArray: function isArray(arr) {
			return Array.isArray(arr);
		},
		webRTCAdapter: deps && deps.adapter || adapter,
		httpAPICall: function httpAPICall(url, options) {
			var fetchOptions = {
				method: options.verb,
				headers: {
					'Accept': 'application/json, text/plain, */*'
				},
				cache: 'no-cache'
			};
			if (options.verb === "POST") {
				fetchOptions.headers['Content-Type'] = 'application/json';
			}
			if (typeof options.withCredentials !== 'undefined') {
				fetchOptions.credentials = options.withCredentials === true ? 'include' : options.withCredentials ? options.withCredentials : 'omit';
			}
			if (options.body) {
				fetchOptions.body = JSON.stringify(options.body);
			}

			var fetching = f(url, fetchOptions)["catch"](function (error) {
				return p.reject({ message: 'Probably a network error, is the server down?', error: error });
			});

			/*
    * fetch() does not natively support timeouts.
    * Work around this by starting a timeout manually, and racing it agains the fetch() to see which thing resolves first.
    */

			if (options.timeout) {
				var timeout = new p(function (resolve, reject) {
					var timerId = setTimeout(function () {
						clearTimeout(timerId);
						return reject({ message: 'Request timed out', timeout: options.timeout });
					}, options.timeout);
				});
				fetching = p.race([fetching, timeout]);
			}

			fetching.then(function (response) {
				if (response.ok) {
					if (typeof options.success === typeof Janus.noop) {
						return response.json().then(function (parsed) {
							try {
								options.success(parsed);
							} catch (error) {
								Janus.error('Unhandled httpAPICall success callback error', error);
							}
						}, function (error) {
							return p.reject({ message: 'Failed to parse response body', error: error, response: response });
						});
					}
				} else {
					return p.reject({ message: 'API call failed', response: response });
				}
			})["catch"](function (error) {
				if (typeof options.error === typeof Janus.noop) {
					options.error(error.message || '<< internal error >>', error);
				}
			});

			return fetching;
		}
	};
};

Janus.useOldDependencies = function (deps) {
	var jq = deps && deps.jQuery || jQuery;
	var socketCls = deps && deps.WebSocket || WebSocket;
	return {
		newWebSocket: function newWebSocket(server, proto) {
			return new socketCls(server, proto);
		},
		isArray: function isArray(arr) {
			return jq.isArray(arr);
		},
		extension: deps && deps.extension || defaultExtension,
		webRTCAdapter: deps && deps.adapter || adapter,
		httpAPICall: function httpAPICall(url, options) {
			var payload = typeof options.body !== 'undefined' ? {
				contentType: 'application/json',
				data: JSON.stringify(options.body)
			} : {};
			var credentials = typeof options.withCredentials !== 'undefined' ? { xhrFields: { withCredentials: options.withCredentials } } : {};

			return jq.ajax(jq.extend(payload, credentials, {
				url: url,
				type: options.verb,
				cache: false,
				dataType: 'json',
				async: options.async,
				timeout: options.timeout,
				success: function success(result) {
					if (typeof options.success === typeof Janus.noop) {
						options.success(result);
					}
				},
				error: function error(xhr, status, err) {
					if (typeof options.error === typeof Janus.noop) {
						options.error(status, err);
					}
				}
			}));
		}
	};
};

// Helper function to convert a deprecated media object to a tracks array
Janus.mediaToTracks = function (media) {
	var tracks = [];
	if (!media) {
		// Default is bidirectional audio and video, using default devices
		tracks.push({ type: 'audio', capture: true, recv: true });
		tracks.push({ type: 'video', capture: true, recv: true });
	} else {
		if (!media.keepAudio && media.audio !== false && (typeof media.audio === 'undefined' || media.audio || media.audioSend || media.audioRecv || media.addAudio || media.replaceAudio || media.removeAudio)) {
			// We may need an audio track
			var track = { type: 'audio' };
			if (media.removeAudio) {
				track.remove = true;
			} else {
				if (media.addAudio) track.add = true;else if (media.replaceAudio) track.replace = true;
				// Check if we need to capture an audio device
				if (media.audioSend !== false) track.capture = media.audio || true;
				// Check if we need to receive audio
				if (media.audioRecv !== false) track.recv = true;
			}
			// Add an audio track if needed
			if (track.remove || track.capture || track.recv) tracks.push(track);
		}
		if (!media.keepVideo && media.video !== false && (typeof media.video === 'undefined' || media.video || media.videoSend || media.videoRecv || media.addVideo || media.replaceVideo || media.removeVideo)) {
			// We may need a video track
			var track = { type: 'video' };
			if (media.removeVideo) {
				track.remove = true;
			} else {
				if (media.addVideo) track.add = true;else if (media.replaceVideo) track.replace = true;
				// Check if we need to capture a video device
				if (media.videoSend !== false) {
					track.capture = media.video || true;
					if (['screen', 'window', 'desktop'].includes(track.capture)) {
						// Change the type to 'screen'
						track.type = 'screen';
						track.capture = { video: {} };
						// Check if there's constraints
						if (media.screenshareFrameRate) track.capture.frameRate = media.screenshareFrameRate;
						if (media.screenshareHeight) track.capture.height = media.screenshareHeight;
						if (media.screenshareWidth) track.capture.width = media.screenshareWidth;
					}
				}
				// Check if we need to receive video
				if (media.videoRecv !== false) track.recv = true;
			}
			// Add a video track if needed
			if (track.remove || track.capture || track.recv) tracks.push(track);
		}
		if (media.data) {
			// We need a data channel
			tracks.push({ type: 'data' });
		}
	}
	// Done
	return tracks;
};

// Helper function to convert a track object to a set of constraints
Janus.trackConstraints = function (track) {
	var constraints = {};
	if (!track || !track.capture) return constraints;
	if (track.type === 'audio') {
		// Just put the capture part in the constraints
		constraints.audio = track.capture;
		// MY_CODE: VIDEO DEFAULTS
	} else if (track.type === 'video') {
			// Check if one of the keywords was passed
			if ((track.simulcast || track.svc) && track.capture === true) track.capture = 'hires';
			if (track.capture === true || typeof track.capture === 'object') {
				// Use the provided capture object as video constraint
				constraints.video = track.capture;
			} else {
				var width = 0;
				var height = 0;
				var maxHeight = 0;
				if (media.video === 'lowres') {
					// Small resolution, 4:3
					height = 240;
					maxHeight = 240;
					width = 320;
				} else if (media.video === 'lowres-16:9') {
					// Small resolution, 16:9
					height = 180;
					maxHeight = 180;
					width = 320;
				} else if (media.video === 'hires' || media.video === 'hires-16:9' || media.video === 'hdres') {
					// High(HD) resolution is only 16:9
					height = 720;
					maxHeight = 720;
					width = 1280;
				} else if (media.video === 'fhdres') {
					// Full HD resolution is only 16:9
					height = 1080;
					maxHeight = 1080;
					width = 1920;
				} else if (media.video === '4kres') {
					// 4K resolution is only 16:9
					height = 2160;
					maxHeight = 2160;
					width = 3840;
				} else if (media.video === 'stdres') {
					// Normal resolution, 4:3
					height = 480;
					maxHeight = 480;
					width = 640;
				} else if (media.video === 'stdres-16:9') {
					// Normal resolution, 16:9
					height = 360;
					maxHeight = 360;
					width = 640;
				} else if (media.video === 'ios-res') {
					// Normal resolution, 9:16
					height = 480;
					maxHeight = 480;
					width = 640;
				} else if (media.video === 'android-stdres') {
					// Normal resolution
					height = 396;
					maxHeight = 396;
					width = 704;
				} else if (media.video === 'android-hires') {
					// Normal resolution
					height = 504;
					maxHeight = 504;
					width = 896;
				} else if (media.video === 'windows-res') {
					height = 480;
					maxHeight = 480;
					width = 853;
				} else if (media.video === 'mac-res') {
					height = 480;
					maxHeight = 480;
					width = 853;
				} else {
					Janus.log("Default video setting is stdres 4:3");
					height = 480;
					maxHeight = 480;
					width = 640;
				}
				constraints.video = {
					width: { ideal: width },
					height: { ideal: height }
				};
			}
		} else if (track.type === 'screen') {
			// Use the provided capture object as video constraint
			constraints.video = track.capture;
		}
	return constraints;
};

Janus.noop = function () {};

Janus.dataChanDefaultLabel = "JanusDataChannel";

// Note: in the future we may want to change this, e.g., as was
// attempted in https://github.com/meetecho/janus-gateway/issues/1670
Janus.endOfCandidates = null;

// Stop all tracks from a given stream
Janus.stopAllTracks = function (stream) {
	try {
		// Try a MediaStreamTrack.stop() for each track
		var tracks = stream.getTracks();
		var _iteratorNormalCompletion = true;
		var _didIteratorError = false;
		var _iteratorError = undefined;

		try {
			for (var _iterator = tracks[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
				var mst = _step.value;

				Janus.log(mst);
				if (mst && mst.dontStop !== true) {
					mst.stop();
				}
			}
		} catch (err) {
			_didIteratorError = true;
			_iteratorError = err;
		} finally {
			try {
				if (!_iteratorNormalCompletion && _iterator["return"]) {
					_iterator["return"]();
				}
			} finally {
				if (_didIteratorError) {
					throw _iteratorError;
				}
			}
		}
	} catch (e) {
		// Do nothing if this fails
	}
};

// Initialization
Janus.init = function (options) {
	options = options || {};
	options.callback = typeof options.callback == "function" ? options.callback : Janus.noop;
	if (Janus.initDone) {
		// Already initialized
		options.callback();
	} else {
		var _iteratorNormalCompletion2;

		var _didIteratorError2;

		var _iteratorError2;

		var _iterator2, _step2;

		var _iteratorNormalCompletion3;

		var _didIteratorError3;

		var _iteratorError3;

		var _iterator3, _step3;

		(function () {
			if (typeof console.log == "undefined") {
				console.log = function () {};
			}
			// Console logging (all debugging disabled by default)
			Janus.trace = Janus.noop;
			Janus.debug = Janus.noop;
			Janus.vdebug = Janus.noop;
			Janus.log = Janus.noop;
			Janus.warn = Janus.noop;
			Janus.error = Janus.noop;
			if (options.debug === true || options.debug === "all") {
				// Enable all debugging levels
				Janus.trace = console.trace.bind(console);
				Janus.debug = console.debug.bind(console);
				Janus.vdebug = console.debug.bind(console);
				Janus.log = console.log.bind(console);
				Janus.warn = console.warn.bind(console);
				Janus.error = console.error.bind(console);
			} else if (Array.isArray(options.debug)) {
				_iteratorNormalCompletion2 = true;
				_didIteratorError2 = false;
				_iteratorError2 = undefined;

				try {
					for (_iterator2 = options.debug[Symbol.iterator](); !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
						var d = _step2.value;

						switch (d) {
							case "trace":
								Janus.trace = console.trace.bind(console);
								break;
							case "debug":
								Janus.debug = console.debug.bind(console);
								break;
							case "vdebug":
								Janus.vdebug = console.debug.bind(console);
								break;
							case "log":
								Janus.log = console.log.bind(console);
								break;
							case "warn":
								Janus.warn = console.warn.bind(console);
								break;
							case "error":
								Janus.error = console.error.bind(console);
								break;
							default:
								console.error("Unknown debugging option '" + d + "' (supported: 'trace', 'debug', 'vdebug', 'log', warn', 'error')");
								break;
						}
					}
				} catch (err) {
					_didIteratorError2 = true;
					_iteratorError2 = err;
				} finally {
					try {
						if (!_iteratorNormalCompletion2 && _iterator2["return"]) {
							_iterator2["return"]();
						}
					} finally {
						if (_didIteratorError2) {
							throw _iteratorError2;
						}
					}
				}
			}
			Janus.log("Initializing library");

			var usedDependencies = options.dependencies || Janus.useDefaultDependencies();
			Janus.isArray = usedDependencies.isArray;
			Janus.webRTCAdapter = usedDependencies.webRTCAdapter;
			Janus.httpAPICall = usedDependencies.httpAPICall;
			Janus.newWebSocket = usedDependencies.newWebSocket;
			Janus.extension = usedDependencies.extension;
			Janus.extension.init();

			// Helper method to enumerate devices
			Janus.listDevices = function (callback, config) {
				callback = typeof callback == "function" ? callback : Janus.noop;
				if (!config) config = { audio: true, video: true };
				if (Janus.isGetUserMediaAvailable()) {
					navigator.mediaDevices.getUserMedia(config).then(function (stream) {
						navigator.mediaDevices.enumerateDevices().then(function (devices) {
							Janus.debug(devices);
							callback(devices);
							// Get rid of the now useless stream
							Janus.stopAllTracks(stream);
						});
					})["catch"](function (err) {
						Janus.error(err);
						callback([]);
					});
				} else {
					Janus.warn("navigator.mediaDevices unavailable");
					callback([]);
				}
			};
			// Helper methods to attach/reattach a stream to a video element (previously part of adapter.js)
			Janus.attachMediaStream = function (element, stream) {
				try {
					element.srcObject = stream;
				} catch (e) {
					try {
						element.src = URL.createObjectURL(stream);
					} catch (e) {
						Janus.error("Error attaching stream to element", e);
					}
				}
			};
			Janus.reattachMediaStream = function (to, from) {
				try {
					to.srcObject = from.srcObject;
				} catch (e) {
					try {
						to.src = from.src;
					} catch (e) {
						Janus.error("Error reattaching stream to element", e);
					}
				}
			};
			// Detect tab close: make sure we don't loose existing onbeforeunload handlers
			// (note: for iOS we need to subscribe to a different event, 'pagehide', see
			// https://gist.github.com/thehunmonkgroup/6bee8941a49b86be31a787fe8f4b8cfe)
			var iOS = ['iPad', 'iPhone', 'iPod'].indexOf(navigator.platform) >= 0;
			var eventName = iOS ? 'pagehide' : 'beforeunload';
			var oldOBF = window["on" + eventName];
			window.addEventListener(eventName, function () {
				Janus.log("Closing window");
				for (var s in Janus.sessions) {
					if (Janus.sessions[s] && Janus.sessions[s].destroyOnUnload) {
						Janus.log("Destroying session " + s);
						Janus.sessions[s].destroy({ unload: true, notifyDestroyed: false });
					}
				}
				if (oldOBF && typeof oldOBF == "function") {
					oldOBF();
				}
			});
			// If this is a Safari Technology Preview, check if VP8 is supported
			Janus.safariVp8 = false;
			if (Janus.webRTCAdapter.browserDetails.browser === 'safari' && Janus.webRTCAdapter.browserDetails.version >= 605) {
				// Let's see if RTCRtpSender.getCapabilities() is there
				if (RTCRtpSender && RTCRtpSender.getCapabilities && RTCRtpSender.getCapabilities("video") && RTCRtpSender.getCapabilities("video").codecs && RTCRtpSender.getCapabilities("video").codecs.length) {
					_iteratorNormalCompletion3 = true;
					_didIteratorError3 = false;
					_iteratorError3 = undefined;

					try {
						for (_iterator3 = RTCRtpSender.getCapabilities("video").codecs[Symbol.iterator](); !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
							var codec = _step3.value;

							if (codec && codec.mimeType && codec.mimeType.toLowerCase() === "video/vp8") {
								Janus.safariVp8 = true;
								break;
							}
						}
					} catch (err) {
						_didIteratorError3 = true;
						_iteratorError3 = err;
					} finally {
						try {
							if (!_iteratorNormalCompletion3 && _iterator3["return"]) {
								_iterator3["return"]();
							}
						} finally {
							if (_didIteratorError3) {
								throw _iteratorError3;
							}
						}
					}

					if (Janus.safariVp8) {
						Janus.log("This version of Safari supports VP8");
					} else {
						Janus.warn("This version of Safari does NOT support VP8: if you're using a Technology Preview, " + "try enabling the 'WebRTC VP8 codec' setting in the 'Experimental Features' Develop menu");
					}
				} else {
					(function () {
						// We do it in a very ugly way, as there's no alternative...
						// We create a PeerConnection to see if VP8 is in an offer
						var testpc = new RTCPeerConnection({});
						testpc.createOffer({ offerToReceiveVideo: true }).then(function (offer) {
							Janus.safariVp8 = offer.sdp.indexOf("VP8") !== -1;
							if (Janus.safariVp8) {
								Janus.log("This version of Safari supports VP8");
							} else {
								Janus.warn("This version of Safari does NOT support VP8: if you're using a Technology Preview, " + "try enabling the 'WebRTC VP8 codec' setting in the 'Experimental Features' Develop menu");
							}
							testpc.close();
							testpc = null;
						});
					})();
				}
			}
			Janus.initDone = true;
			options.callback();
		})();
	}
};

// Helper method to check whether WebRTC is supported by this browser
Janus.isWebrtcSupported = function () {
	return !!window.RTCPeerConnection;
};
// Helper method to check whether devices can be accessed by this browser (e.g., not possible via plain HTTP)
Janus.isGetUserMediaAvailable = function () {
	return navigator.mediaDevices && navigator.mediaDevices.getUserMedia;
};

// Helper method to create random identifiers (e.g., transaction)
Janus.randomString = function (len) {
	var charSet = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789';
	var randomString = '';
	for (var i = 0; i < len; i++) {
		var randomPoz = Math.floor(Math.random() * charSet.length);
		randomString += charSet.charAt(randomPoz);
	}
	return randomString;
};

function Janus(gatewayCallbacks) {
	gatewayCallbacks = gatewayCallbacks || {};
	gatewayCallbacks.success = typeof gatewayCallbacks.success == "function" ? gatewayCallbacks.success : Janus.noop;
	gatewayCallbacks.error = typeof gatewayCallbacks.error == "function" ? gatewayCallbacks.error : Janus.noop;
	gatewayCallbacks.destroyed = typeof gatewayCallbacks.destroyed == "function" ? gatewayCallbacks.destroyed : Janus.noop;
	if (!Janus.initDone) {
		gatewayCallbacks.error("Library not initialized");
		return {};
	}
	if (!Janus.isWebrtcSupported()) {
		gatewayCallbacks.error("WebRTC not supported by this browser");
		return {};
	}
	Janus.log("Library initialized: " + Janus.initDone);
	if (!gatewayCallbacks.server) {
		gatewayCallbacks.error("Invalid server url");
		return {};
	}
	var websockets = false;
	var ws = null;
	var wsHandlers = {};
	var wsKeepaliveTimeoutId = null;
	var servers = null;
	var serversIndex = 0;
	var server = gatewayCallbacks.server;
	if (Janus.isArray(server)) {
		Janus.log("Multiple servers provided (" + server.length + "), will use the first that works");
		server = null;
		servers = gatewayCallbacks.server;
		Janus.debug(servers);
	} else {
		if (server.indexOf("ws") === 0) {
			websockets = true;
			Janus.log("Using WebSockets to contact Janus: " + server);
		} else {
			websockets = false;
			Janus.log("Using REST API to contact Janus: " + server);
		}
	}
	var iceServers = gatewayCallbacks.iceServers || [{ urls: "stun:stun.l.google.com:19302" }];
	var iceTransportPolicy = gatewayCallbacks.iceTransportPolicy;
	var bundlePolicy = gatewayCallbacks.bundlePolicy;
	// Whether we should enable the withCredentials flag for XHR requests
	var withCredentials = false;
	if (typeof gatewayCallbacks.withCredentials !== 'undefined' && gatewayCallbacks.withCredentials !== null) withCredentials = gatewayCallbacks.withCredentials === true;
	// Optional max events
	var maxev = 10;
	if (typeof gatewayCallbacks.max_poll_events !== 'undefined' && gatewayCallbacks.max_poll_events !== null) maxev = gatewayCallbacks.max_poll_events;
	if (maxev < 1) maxev = 1;
	// Token to use (only if the token based authentication mechanism is enabled)
	var token = null;
	if (typeof gatewayCallbacks.token !== 'undefined' && gatewayCallbacks.token !== null) token = gatewayCallbacks.token;
	// API secret to use (only if the shared API secret is enabled)
	var apisecret = null;
	if (typeof gatewayCallbacks.apisecret !== 'undefined' && gatewayCallbacks.apisecret !== null) apisecret = gatewayCallbacks.apisecret;
	// Whether we should destroy this session when onbeforeunload is called
	this.destroyOnUnload = true;
	if (typeof gatewayCallbacks.destroyOnUnload !== 'undefined' && gatewayCallbacks.destroyOnUnload !== null) this.destroyOnUnload = gatewayCallbacks.destroyOnUnload === true;
	// Some timeout-related values
	var keepAlivePeriod = 25000;
	if (typeof gatewayCallbacks.keepAlivePeriod !== 'undefined' && gatewayCallbacks.keepAlivePeriod !== null) keepAlivePeriod = gatewayCallbacks.keepAlivePeriod;
	if (isNaN(keepAlivePeriod)) keepAlivePeriod = 25000;
	var longPollTimeout = 60000;
	if (typeof gatewayCallbacks.longPollTimeout !== 'undefined' && gatewayCallbacks.longPollTimeout !== null) longPollTimeout = gatewayCallbacks.longPollTimeout;
	if (isNaN(longPollTimeout)) longPollTimeout = 60000;

	// overrides for default maxBitrate values for simulcasting
	function getMaxBitrates(simulcastMaxBitrates) {
		var maxBitrates = {
			high: 900000,
			medium: 300000,
			low: 100000
		};

		if (typeof simulcastMaxBitrates !== 'undefined' && simulcastMaxBitrates !== null) {
			if (simulcastMaxBitrates.high) maxBitrates.high = simulcastMaxBitrates.high;
			if (simulcastMaxBitrates.medium) maxBitrates.medium = simulcastMaxBitrates.medium;
			if (simulcastMaxBitrates.low) maxBitrates.low = simulcastMaxBitrates.low;
		}

		return maxBitrates;
	}

	var connected = false;
	var sessionId = null;
	var pluginHandles = {};
	var that = this;
	var retries = 0;
	var transactions = {};
	createSession(gatewayCallbacks);

	// Public methods
	this.getServer = function () {
		return server;
	};
	this.isConnected = function () {
		return connected;
	};
	this.reconnect = function (callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		callbacks["reconnect"] = true;
		createSession(callbacks);
	};
	this.getSessionId = function () {
		return sessionId;
	};
	this.getInfo = function (callbacks) {
		getInfo(callbacks);
	};
	this.destroy = function (callbacks) {
		destroySession(callbacks);
	};
	this.attach = function (callbacks) {
		createHandle(callbacks);
	};

	function eventHandler() {
		if (sessionId == null) return;
		Janus.debug('Long poll...');
		if (!connected) {
			Janus.warn("Is the server down? (connected=false)");
			return;
		}
		var longpoll = server + "/" + sessionId + "?rid=" + new Date().getTime();
		if (maxev) longpoll = longpoll + "&maxev=" + maxev;
		if (token) longpoll = longpoll + "&token=" + encodeURIComponent(token);
		if (apisecret) longpoll = longpoll + "&apisecret=" + encodeURIComponent(apisecret);
		Janus.httpAPICall(longpoll, {
			verb: 'GET',
			withCredentials: withCredentials,
			success: handleEvent,
			timeout: longPollTimeout,
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown);
				retries++;
				if (retries > 3) {
					// Did we just lose the server? :-(
					connected = false;
					gatewayCallbacks.error("Lost connection to the server (is it down?)");
					return;
				}
				eventHandler();
			}
		});
	}

	// Private event handler: this will trigger plugin callbacks, if set
	function handleEvent(json, skipTimeout) {
		retries = 0;
		if (!websockets && typeof sessionId !== 'undefined' && sessionId !== null && skipTimeout !== true) eventHandler();
		if (!websockets && Janus.isArray(json)) {
			// We got an array: it means we passed a maxev > 1, iterate on all objects
			for (var i = 0; i < json.length; i++) {
				handleEvent(json[i], true);
			}
			return;
		}
		if (json["janus"] === "keepalive") {
			// Nothing happened
			Janus.vdebug("Got a keepalive on session " + sessionId);
			return;
		} else if (json["janus"] === "server_info") {
			// Just info on the Janus instance
			Janus.debug("Got info on the Janus instance");
			Janus.debug(json);
			var transaction = json["transaction"];
			if (transaction) {
				var reportSuccess = transactions[transaction];
				if (reportSuccess) reportSuccess(json);
				delete transactions[transaction];
			}
			return;
		} else if (json["janus"] === "ack") {
			// Just an ack, we can probably ignore
			Janus.debug("Got an ack on session " + sessionId);
			Janus.debug(json);
			var transaction = json["transaction"];
			if (transaction) {
				var reportSuccess = transactions[transaction];
				if (reportSuccess) reportSuccess(json);
				delete transactions[transaction];
			}
			return;
		} else if (json["janus"] === "success") {
			// Success!
			Janus.debug("Got a success on session " + sessionId);
			Janus.debug(json);
			var transaction = json["transaction"];
			if (transaction) {
				var reportSuccess = transactions[transaction];
				if (reportSuccess) reportSuccess(json);
				delete transactions[transaction];
			}
			return;
		} else if (json["janus"] === "trickle") {
			// We got a trickle candidate from Janus
			var sender = json["sender"];
			if (!sender) {
				Janus.warn("Missing sender...");
				return;
			}
			var pluginHandle = pluginHandles[sender];
			if (!pluginHandle) {
				Janus.debug("This handle is not attached to this session");
				return;
			}
			var candidate = json["candidate"];
			Janus.debug("Got a trickled candidate on session " + sessionId);
			Janus.debug(candidate);
			var config = pluginHandle.webrtcStuff;
			if (config.pc && config.remoteSdp) {
				// Add candidate right now
				Janus.debug("Adding remote candidate:", candidate);
				if (!candidate || candidate.completed === true) {
					// end-of-candidates
					config.pc.addIceCandidate(Janus.endOfCandidates);
				} else {
					// New candidate
					config.pc.addIceCandidate(candidate);
				}
			} else {
				// We didn't do setRemoteDescription (trickle got here before the offer?)
				Janus.debug("We didn't do setRemoteDescription (trickle got here before the offer?), caching candidate");
				if (!config.candidates) config.candidates = [];
				config.candidates.push(candidate);
				Janus.debug(config.candidates);
			}
		} else if (json["janus"] === "webrtcup") {
			// The PeerConnection with the server is up! Notify this
			Janus.debug("Got a webrtcup event on session " + sessionId);
			Janus.debug(json);
			var sender = json["sender"];
			if (!sender) {
				Janus.warn("Missing sender...");
				return;
			}
			var pluginHandle = pluginHandles[sender];
			if (!pluginHandle) {
				Janus.debug("This handle is not attached to this session");
				return;
			}
			pluginHandle.webrtcState(true);
			return;
		} else if (json["janus"] === "hangup") {
			// A plugin asked the core to hangup a PeerConnection on one of our handles
			Janus.debug("Got a hangup event on session " + sessionId);
			Janus.debug(json);
			var sender = json["sender"];
			if (!sender) {
				Janus.warn("Missing sender...");
				return;
			}
			var pluginHandle = pluginHandles[sender];
			if (!pluginHandle) {
				Janus.debug("This handle is not attached to this session");
				return;
			}
			pluginHandle.webrtcState(false, json["reason"]);
			pluginHandle.hangup();
		} else if (json["janus"] === "detached") {
			// A plugin asked the core to detach one of our handles
			Janus.debug("Got a detached event on session " + sessionId);
			Janus.debug(json);
			var sender = json["sender"];
			if (!sender) {
				Janus.warn("Missing sender...");
				return;
			}
			var pluginHandle = pluginHandles[sender];
			if (!pluginHandle) {
				// Don't warn here because destroyHandle causes this situation.
				return;
			}
			pluginHandle.ondetached();
			pluginHandle.detach();
		} else if (json["janus"] === "media") {
			// Media started/stopped flowing
			Janus.debug("Got a media event on session " + sessionId);
			Janus.debug(json);
			var sender = json["sender"];
			if (!sender) {
				Janus.warn("Missing sender...");
				return;
			}
			var pluginHandle = pluginHandles[sender];
			if (!pluginHandle) {
				Janus.debug("This handle is not attached to this session");
				return;
			}
			pluginHandle.mediaState(json["type"], json["receiving"], json["mid"]);
		} else if (json["janus"] === "slowlink") {
			Janus.debug("Got a slowlink event on session " + sessionId);
			Janus.debug(json);
			// Trouble uplink or downlink
			var sender = json["sender"];
			if (!sender) {
				Janus.warn("Missing sender...");
				return;
			}
			var pluginHandle = pluginHandles[sender];
			if (!pluginHandle) {
				Janus.debug("This handle is not attached to this session");
				return;
			}
			pluginHandle.slowLink(json["uplink"], json["lost"], json["mid"]);
		} else if (json["janus"] === "error") {
			// Oops, something wrong happened
			Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
			Janus.debug(json);
			var transaction = json["transaction"];
			if (transaction) {
				var reportSuccess = transactions[transaction];
				if (reportSuccess) {
					reportSuccess(json);
				}
				delete transactions[transaction];
			}
			return;
		} else if (json["janus"] === "event") {
			Janus.debug("Got a plugin event on session " + sessionId);
			Janus.debug(json);
			var sender = json["sender"];
			if (!sender) {
				Janus.warn("Missing sender...");
				return;
			}
			var plugindata = json["plugindata"];
			if (!plugindata) {
				Janus.warn("Missing plugindata...");
				return;
			}
			Janus.debug("  -- Event is coming from " + sender + " (" + plugindata["plugin"] + ")");
			var data = plugindata["data"];
			Janus.debug(data);
			var pluginHandle = pluginHandles[sender];
			if (!pluginHandle) {
				Janus.warn("This handle is not attached to this session");
				return;
			}
			var jsep = json["jsep"];
			if (jsep) {
				Janus.debug("Handling SDP as well...");
				Janus.debug(jsep);
			}
			var callback = pluginHandle.onmessage;
			if (callback) {
				Janus.debug("Notifying application...");
				// Send to callback specified when attaching plugin handle
				callback(data, jsep);
			} else {
				// Send to generic callback (?)
				Janus.debug("No provided notification callback");
			}
		} else if (json["janus"] === "timeout") {
			Janus.error("Timeout on session " + sessionId);
			Janus.debug(json);
			if (websockets) {
				ws.close(3504, "Gateway timeout");
			}
			return;
		} else {
			Janus.warn("Unknown message/event  '" + json["janus"] + "' on session " + sessionId);
			Janus.debug(json);
		}
	}

	// Private helper to send keep-alive messages on WebSockets
	function keepAlive() {
		if (!server || !websockets || !connected) return;
		wsKeepaliveTimeoutId = setTimeout(keepAlive, keepAlivePeriod);
		var request = { "janus": "keepalive", "session_id": sessionId, "transaction": Janus.randomString(12) };
		if (token) request["token"] = token;
		if (apisecret) request["apisecret"] = apisecret;
		ws.send(JSON.stringify(request));
	}

	// Private method to create a session
	function createSession(callbacks) {
		var transaction = Janus.randomString(12);
		var request = { "janus": "create", "transaction": transaction };
		if (callbacks["reconnect"]) {
			// We're reconnecting, claim the session
			connected = false;
			request["janus"] = "claim";
			request["session_id"] = sessionId;
			// If we were using websockets, ignore the old connection
			if (ws) {
				ws.onopen = null;
				ws.onerror = null;
				ws.onclose = null;
				if (wsKeepaliveTimeoutId) {
					clearTimeout(wsKeepaliveTimeoutId);
					wsKeepaliveTimeoutId = null;
				}
			}
		}
		if (token) request["token"] = token;
		if (apisecret) request["apisecret"] = apisecret;
		if (!server && Janus.isArray(servers)) {
			// We still need to find a working server from the list we were given
			server = servers[serversIndex];
			if (server.indexOf("ws") === 0) {
				websockets = true;
				Janus.log("Server #" + (serversIndex + 1) + ": trying WebSockets to contact Janus (" + server + ")");
			} else {
				websockets = false;
				Janus.log("Server #" + (serversIndex + 1) + ": trying REST API to contact Janus (" + server + ")");
			}
		}
		if (websockets) {
			ws = Janus.newWebSocket(server, 'janus-protocol');
			wsHandlers = {
				'error': function error() {
					Janus.error("Error connecting to the Janus WebSockets server... " + server);
					if (Janus.isArray(servers) && !callbacks["reconnect"]) {
						serversIndex++;
						if (serversIndex === servers.length) {
							// We tried all the servers the user gave us and they all failed
							callbacks.error("Error connecting to any of the provided Janus servers: Is the server down?");
							return;
						}
						// Let's try the next server
						server = null;
						setTimeout(function () {
							createSession(callbacks);
						}, 200);
						return;
					}
					callbacks.error("Error connecting to the Janus WebSockets server: Is the server down?");
				},

				'open': function open() {
					// We need to be notified about the success
					transactions[transaction] = function (json) {
						Janus.debug(json);
						if (json["janus"] !== "success") {
							Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
							callbacks.error(json["error"].reason);
							return;
						}
						wsKeepaliveTimeoutId = setTimeout(keepAlive, keepAlivePeriod);
						connected = true;
						sessionId = json["session_id"] ? json["session_id"] : json.data["id"];
						if (callbacks["reconnect"]) {
							Janus.log("Claimed session: " + sessionId);
						} else {
							Janus.log("Created session: " + sessionId);
						}
						Janus.sessions[sessionId] = that;
						callbacks.success();
					};
					ws.send(JSON.stringify(request));
				},

				'message': function message(event) {
					handleEvent(JSON.parse(event.data));
				},

				'close': function close() {
					if (!server || !connected) {
						return;
					}
					connected = false;
					// FIXME What if this is called when the page is closed?
					gatewayCallbacks.error("Lost connection to the server (is it down?)");
				}
			};

			for (var eventName in wsHandlers) {
				ws.addEventListener(eventName, wsHandlers[eventName]);
			}

			return;
		}
		Janus.httpAPICall(server, {
			verb: 'POST',
			withCredentials: withCredentials,
			body: request,
			success: function success(json) {
				Janus.debug(json);
				if (json["janus"] !== "success") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
					callbacks.error(json["error"].reason);
					return;
				}
				connected = true;
				sessionId = json["session_id"] ? json["session_id"] : json.data["id"];
				if (callbacks["reconnect"]) {
					Janus.log("Claimed session: " + sessionId);
				} else {
					Janus.log("Created session: " + sessionId);
				}
				Janus.sessions[sessionId] = that;
				eventHandler();
				callbacks.success();
			},
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown); // FIXME
				if (Janus.isArray(servers) && !callbacks["reconnect"]) {
					serversIndex++;
					if (serversIndex === servers.length) {
						// We tried all the servers the user gave us and they all failed
						callbacks.error("Error connecting to any of the provided Janus servers: Is the server down?");
						return;
					}
					// Let's try the next server
					server = null;
					setTimeout(function () {
						createSession(callbacks);
					}, 200);
					return;
				}
				if (errorThrown === "") callbacks.error(textStatus + ": Is the server down?");else if (errorThrown && errorThrown.error) callbacks.error(textStatus + ": " + errorThrown.error.message);else callbacks.error(textStatus + ": " + errorThrown);
			}
		});
	}

	// Private method to get info on the server
	function getInfo(callbacks) {
		callbacks = callbacks || {};
		// FIXME This method triggers a success even when we fail
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		Janus.log("Getting info on Janus instance");
		if (!connected) {
			Janus.warn("Is the server down? (connected=false)");
			callbacks.error("Is the server down? (connected=false)");
			return;
		}
		// We just need to send an "info" request
		var transaction = Janus.randomString(12);
		var request = { "janus": "info", "transaction": transaction };
		if (token) request["token"] = token;
		if (apisecret) request["apisecret"] = apisecret;
		if (websockets) {
			transactions[transaction] = function (json) {
				Janus.log("Server info:");
				Janus.debug(json);
				if (json["janus"] !== "server_info") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
				}
				callbacks.success(json);
			};
			ws.send(JSON.stringify(request));
			return;
		}
		Janus.httpAPICall(server, {
			verb: 'POST',
			withCredentials: withCredentials,
			body: request,
			success: function success(json) {
				Janus.log("Server info:");
				Janus.debug(json);
				if (json["janus"] !== "server_info") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
				}
				callbacks.success(json);
			},
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown); // FIXME
				if (errorThrown === "") callbacks.error(textStatus + ": Is the server down?");else callbacks.error(textStatus + ": " + errorThrown);
			}
		});
	}

	// Private method to destroy a session
	function destroySession(callbacks) {
		callbacks = callbacks || {};
		// FIXME This method triggers a success even when we fail
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		var unload = callbacks.unload === true;
		var notifyDestroyed = true;
		if (typeof callbacks.notifyDestroyed !== 'undefined' && callbacks.notifyDestroyed !== null) notifyDestroyed = callbacks.notifyDestroyed === true;
		var cleanupHandles = callbacks.cleanupHandles === true;
		Janus.log("Destroying session " + sessionId + " (unload=" + unload + ")");
		if (!sessionId) {
			Janus.warn("No session to destroy");
			callbacks.success();
			if (notifyDestroyed) gatewayCallbacks.destroyed();
			return;
		}
		if (cleanupHandles) {
			for (var handleId in pluginHandles) {
				destroyHandle(handleId, { noRequest: true });
			}
		}
		if (!connected) {
			Janus.warn("Is the server down? (connected=false)");
			sessionId = null;
			callbacks.success();
			return;
		}
		// No need to destroy all handles first, Janus will do that itself
		var request = { "janus": "destroy", "transaction": Janus.randomString(12) };
		if (token) request["token"] = token;
		if (apisecret) request["apisecret"] = apisecret;
		if (unload) {
			// We're unloading the page: use sendBeacon for HTTP instead,
			// or just close the WebSocket connection if we're using that
			if (websockets) {
				ws.onclose = null;
				ws.close();
				ws = null;
			} else {
				navigator.sendBeacon(server + "/" + sessionId, JSON.stringify(request));
			}
			Janus.log("Destroyed session:");
			sessionId = null;
			connected = false;
			callbacks.success();
			if (notifyDestroyed) gatewayCallbacks.destroyed();
			return;
		}
		if (websockets) {
			var _ret3 = (function () {
				request["session_id"] = sessionId;

				var unbindWebSocket = function unbindWebSocket() {
					for (var eventName in wsHandlers) {
						ws.removeEventListener(eventName, wsHandlers[eventName]);
					}
					ws.removeEventListener('message', onUnbindMessage);
					ws.removeEventListener('error', onUnbindError);
					if (wsKeepaliveTimeoutId) {
						clearTimeout(wsKeepaliveTimeoutId);
					}
					ws.close();
				};

				var onUnbindMessage = function onUnbindMessage(event) {
					var data = JSON.parse(event.data);
					if (data.session_id == request.session_id && data.transaction == request.transaction) {
						unbindWebSocket();
						callbacks.success();
						if (notifyDestroyed) gatewayCallbacks.destroyed();
					}
				};
				var onUnbindError = function onUnbindError() {
					unbindWebSocket();
					callbacks.error("Failed to destroy the server: Is the server down?");
					if (notifyDestroyed) gatewayCallbacks.destroyed();
				};

				ws.addEventListener('message', onUnbindMessage);
				ws.addEventListener('error', onUnbindError);

				if (ws.readyState === 1) {
					ws.send(JSON.stringify(request));
				} else {
					onUnbindError();
				}

				return {
					v: undefined
				};
			})();

			if (typeof _ret3 === "object") return _ret3.v;
		}
		Janus.httpAPICall(server + "/" + sessionId, {
			verb: 'POST',
			withCredentials: withCredentials,
			body: request,
			success: function success(json) {
				Janus.log("Destroyed session:");
				Janus.debug(json);
				sessionId = null;
				connected = false;
				if (json["janus"] !== "success") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
				}
				callbacks.success();
				if (notifyDestroyed) gatewayCallbacks.destroyed();
			},
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown); // FIXME
				// Reset everything anyway
				sessionId = null;
				connected = false;
				callbacks.success();
				if (notifyDestroyed) gatewayCallbacks.destroyed();
			}
		});
	}

	// Private method to create a plugin handle
	function createHandle(callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		callbacks.dataChannelOptions = callbacks.dataChannelOptions || { ordered: true };
		callbacks.consentDialog = typeof callbacks.consentDialog == "function" ? callbacks.consentDialog : Janus.noop;
		callbacks.iceState = typeof callbacks.iceState == "function" ? callbacks.iceState : Janus.noop;
		callbacks.mediaState = typeof callbacks.mediaState == "function" ? callbacks.mediaState : Janus.noop;
		callbacks.webrtcState = typeof callbacks.webrtcState == "function" ? callbacks.webrtcState : Janus.noop;
		callbacks.slowLink = typeof callbacks.slowLink == "function" ? callbacks.slowLink : Janus.noop;
		callbacks.onmessage = typeof callbacks.onmessage == "function" ? callbacks.onmessage : Janus.noop;
		callbacks.onlocaltrack = typeof callbacks.onlocaltrack == "function" ? callbacks.onlocaltrack : Janus.noop;
		callbacks.onremotetrack = typeof callbacks.onremotetrack == "function" ? callbacks.onremotetrack : Janus.noop;
		callbacks.ondata = typeof callbacks.ondata == "function" ? callbacks.ondata : Janus.noop;
		callbacks.ondataopen = typeof callbacks.ondataopen == "function" ? callbacks.ondataopen : Janus.noop;
		callbacks.oncleanup = typeof callbacks.oncleanup == "function" ? callbacks.oncleanup : Janus.noop;
		callbacks.ondetached = typeof callbacks.ondetached == "function" ? callbacks.ondetached : Janus.noop;
		if (!connected) {
			Janus.warn("Is the server down? (connected=false)");
			callbacks.error("Is the server down? (connected=false)");
			return;
		}
		var plugin = callbacks.plugin;
		if (!plugin) {
			Janus.error("Invalid plugin");
			callbacks.error("Invalid plugin");
			return;
		}
		var opaqueId = callbacks.opaqueId;
		var loopIndex = callbacks.loopIndex;
		var handleToken = callbacks.token ? callbacks.token : token;
		var transaction = Janus.randomString(12);
		var request = { "janus": "attach", "plugin": plugin, "opaque_id": opaqueId, "loop_index": loopIndex, "transaction": transaction };
		if (handleToken) request["token"] = handleToken;
		if (apisecret) request["apisecret"] = apisecret;
		if (websockets) {
			transactions[transaction] = function (json) {
				Janus.debug(json);
				if (json["janus"] !== "success") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
					callbacks.error("Ooops: " + json["error"].code + " " + json["error"].reason);
					return;
				}
				var handleId = json.data["id"];
				Janus.log("Created handle: " + handleId);
				var pluginHandle = {
					session: that,
					plugin: plugin,
					id: handleId,
					token: handleToken,
					detached: false,
					webrtcStuff: {
						started: false,
						myStream: null,
						streamExternal: false,
						mySdp: null,
						mediaConstraints: null,
						pc: null,
						dataChannelOptions: callbacks.dataChannelOptions,
						dataChannel: {},
						dtmfSender: null,
						trickle: true,
						iceDone: false,
						bitrate: {}
					},
					getId: function getId() {
						return handleId;
					},
					getPlugin: function getPlugin() {
						return plugin;
					},
					getVolume: function getVolume(mid, result) {
						return _getVolume(handleId, mid, true, result);
					},
					getRemoteVolume: function getRemoteVolume(mid, result) {
						return _getVolume(handleId, mid, true, result);
					},
					getLocalVolume: function getLocalVolume(mid, result) {
						return _getVolume(handleId, mid, false, result);
					},
					isAudioMuted: function isAudioMuted(mid) {
						return isMuted(handleId, mid, false);
					},
					muteAudio: function muteAudio(mid) {
						return mute(handleId, mid, false, true);
					},
					unmuteAudio: function unmuteAudio(mid) {
						return mute(handleId, mid, false, false);
					},
					isVideoMuted: function isVideoMuted(mid) {
						return isMuted(handleId, mid, true);
					},
					muteVideo: function muteVideo(mid) {
						return mute(handleId, mid, true, true);
					},
					unmuteVideo: function unmuteVideo(mid) {
						return mute(handleId, mid, true, false);
					},
					getBitrate: function getBitrate(mid) {
						return _getBitrate(handleId, mid);
					},
					setMaxBitrate: function setMaxBitrate(mid, bitrate) {
						return setBitrate(handleId, mid, bitrate);
					},
					send: function send(callbacks) {
						sendMessage(handleId, callbacks);
					},
					data: function data(callbacks) {
						sendData(handleId, callbacks);
					},
					dtmf: function dtmf(callbacks) {
						sendDtmf(handleId, callbacks);
					},
					consentDialog: callbacks.consentDialog,
					iceState: callbacks.iceState,
					mediaState: callbacks.mediaState,
					webrtcState: callbacks.webrtcState,
					slowLink: callbacks.slowLink,
					onmessage: callbacks.onmessage,
					createOffer: function createOffer(callbacks) {
						prepareWebrtc(handleId, true, callbacks);
					},
					createAnswer: function createAnswer(callbacks) {
						prepareWebrtc(handleId, false, callbacks);
					},
					handleRemoteJsep: function handleRemoteJsep(callbacks) {
						prepareWebrtcPeer(handleId, callbacks);
					},
					replaceTracks: function replaceTracks(callbacks) {
						_replaceTracks(handleId, callbacks);
					},
					getLocalTracks: function getLocalTracks() {
						return _getLocalTracks(handleId);
					},
					getRemoteTracks: function getRemoteTracks() {
						return _getRemoteTracks(handleId);
					},
					onlocaltrack: callbacks.onlocaltrack,
					onremotetrack: callbacks.onremotetrack,
					ondata: callbacks.ondata,
					ondataopen: callbacks.ondataopen,
					oncleanup: callbacks.oncleanup,
					ondetached: callbacks.ondetached,
					hangup: function hangup(sendRequest) {
						cleanupWebrtc(handleId, sendRequest === true);
					},
					detach: function detach(callbacks) {
						destroyHandle(handleId, callbacks);
					}
				};
				pluginHandles[handleId] = pluginHandle;
				callbacks.success(pluginHandle);
			};
			request["session_id"] = sessionId;
			ws.send(JSON.stringify(request));
			return;
		}
		Janus.httpAPICall(server + "/" + sessionId, {
			verb: 'POST',
			withCredentials: withCredentials,
			body: request,
			success: function success(json) {
				Janus.debug(json);
				if (json["janus"] !== "success") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
					callbacks.error("Ooops: " + json["error"].code + " " + json["error"].reason);
					return;
				}
				var handleId = json.data["id"];
				Janus.log("Created handle: " + handleId);
				var pluginHandle = {
					session: that,
					plugin: plugin,
					id: handleId,
					token: handleToken,
					detached: false,
					webrtcStuff: {
						started: false,
						myStream: null,
						streamExternal: false,
						mySdp: null,
						mediaConstraints: null,
						pc: null,
						dataChannelOptions: callbacks.dataChannelOptions,
						dataChannel: {},
						dtmfSender: null,
						trickle: true,
						iceDone: false,
						bitrate: {}
					},
					getId: function getId() {
						return handleId;
					},
					getPlugin: function getPlugin() {
						return plugin;
					},
					getVolume: function getVolume(mid, result) {
						return _getVolume(handleId, mid, true, result);
					},
					getRemoteVolume: function getRemoteVolume(mid, result) {
						return _getVolume(handleId, mid, true, result);
					},
					getLocalVolume: function getLocalVolume(mid, result) {
						return _getVolume(handleId, mid, false, result);
					},
					isAudioMuted: function isAudioMuted(mid) {
						return isMuted(handleId, mid, false);
					},
					muteAudio: function muteAudio(mid) {
						return mute(handleId, mid, false, true);
					},
					unmuteAudio: function unmuteAudio(mid) {
						return mute(handleId, mid, false, false);
					},
					isVideoMuted: function isVideoMuted(mid) {
						return isMuted(handleId, mid, true);
					},
					muteVideo: function muteVideo(mid) {
						return mute(handleId, mid, true, true);
					},
					unmuteVideo: function unmuteVideo(mid) {
						return mute(handleId, mid, true, false);
					},
					getBitrate: function getBitrate(mid) {
						return _getBitrate(handleId, mid);
					},
					setMaxBitrate: function setMaxBitrate(mid, bitrate) {
						return setBitrate(handleId, mid, bitrate);
					},
					send: function send(callbacks) {
						sendMessage(handleId, callbacks);
					},
					data: function data(callbacks) {
						sendData(handleId, callbacks);
					},
					dtmf: function dtmf(callbacks) {
						sendDtmf(handleId, callbacks);
					},
					consentDialog: callbacks.consentDialog,
					iceState: callbacks.iceState,
					mediaState: callbacks.mediaState,
					webrtcState: callbacks.webrtcState,
					slowLink: callbacks.slowLink,
					onmessage: callbacks.onmessage,
					createOffer: function createOffer(callbacks) {
						prepareWebrtc(handleId, true, callbacks);
					},
					createAnswer: function createAnswer(callbacks) {
						prepareWebrtc(handleId, false, callbacks);
					},
					handleRemoteJsep: function handleRemoteJsep(callbacks) {
						prepareWebrtcPeer(handleId, callbacks);
					},
					replaceTracks: function replaceTracks(callbacks) {
						_replaceTracks(handleId, callbacks);
					},
					getLocalTracks: function getLocalTracks() {
						return _getLocalTracks(handleId);
					},
					getRemoteTracks: function getRemoteTracks() {
						return _getRemoteTracks(handleId);
					},
					onlocaltrack: callbacks.onlocaltrack,
					onremotetrack: callbacks.onremotetrack,
					ondata: callbacks.ondata,
					ondataopen: callbacks.ondataopen,
					oncleanup: callbacks.oncleanup,
					ondetached: callbacks.ondetached,
					hangup: function hangup(sendRequest) {
						cleanupWebrtc(handleId, sendRequest === true);
					},
					detach: function detach(callbacks) {
						destroyHandle(handleId, callbacks);
					}
				};
				pluginHandles[handleId] = pluginHandle;
				callbacks.success(pluginHandle);
			},
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown); // FIXME
				if (errorThrown === "") callbacks.error(textStatus + ": Is the server down?");else callbacks.error(textStatus + ": " + errorThrown);
			}
		});
	}

	// Private method to send a message
	function sendMessage(handleId, callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		if (!connected) {
			Janus.warn("Is the server down? (connected=false)");
			callbacks.error("Is the server down? (connected=false)");
			return;
		}
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			callbacks.error("Invalid handle");
			return;
		}
		var message = callbacks.message;
		var jsep = callbacks.jsep;
		var transaction = Janus.randomString(12);
		var request = { "janus": "message", "body": message, "transaction": transaction };
		if (pluginHandle.token) request["token"] = pluginHandle.token;
		if (apisecret) request["apisecret"] = apisecret;
		if (jsep) {
			request.jsep = {
				type: jsep.type,
				sdp: jsep.sdp
			};
			if (jsep.e2ee) request.jsep.e2ee = true;
			if (jsep.rid_order === "hml" || jsep.rid_order === "lmh") request.jsep.rid_order = jsep.rid_order;
			if (jsep.force_relay) request.jsep.force_relay = true;
		}
		Janus.debug("Sending message to plugin (handle=" + handleId + "):");
		Janus.debug(request);
		if (websockets) {
			request["session_id"] = sessionId;
			request["handle_id"] = handleId;
			transactions[transaction] = function (json) {
				Janus.debug("Message sent!");
				Janus.debug(json);
				if (json["janus"] === "success") {
					// We got a success, must have been a synchronous transaction
					var plugindata = json["plugindata"];
					if (!plugindata) {
						Janus.warn("Request succeeded, but missing plugindata...");
						callbacks.success();
						return;
					}
					Janus.log("Synchronous transaction successful (" + plugindata["plugin"] + ")");
					var data = plugindata["data"];
					Janus.debug(data);
					callbacks.success(data);
					return;
				} else if (json["janus"] !== "ack") {
					// Not a success and not an ack, must be an error
					if (json["error"]) {
						Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
						callbacks.error(json["error"].code + " " + json["error"].reason);
					} else {
						Janus.error("Unknown error"); // FIXME
						callbacks.error("Unknown error");
					}
					return;
				}
				// If we got here, the plugin decided to handle the request asynchronously
				callbacks.success();
			};
			ws.send(JSON.stringify(request));
			return;
		}
		Janus.httpAPICall(server + "/" + sessionId + "/" + handleId, {
			verb: 'POST',
			withCredentials: withCredentials,
			body: request,
			success: function success(json) {
				Janus.debug("Message sent!");
				Janus.debug(json);
				if (json["janus"] === "success") {
					// We got a success, must have been a synchronous transaction
					var plugindata = json["plugindata"];
					if (!plugindata) {
						Janus.warn("Request succeeded, but missing plugindata...");
						callbacks.success();
						return;
					}
					Janus.log("Synchronous transaction successful (" + plugindata["plugin"] + ")");
					var data = plugindata["data"];
					Janus.debug(data);
					callbacks.success(data);
					return;
				} else if (json["janus"] !== "ack") {
					// Not a success and not an ack, must be an error
					if (json["error"]) {
						Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
						callbacks.error(json["error"].code + " " + json["error"].reason);
					} else {
						Janus.error("Unknown error"); // FIXME
						callbacks.error("Unknown error");
					}
					return;
				}
				// If we got here, the plugin decided to handle the request asynchronously
				callbacks.success();
			},
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown); // FIXME
				callbacks.error(textStatus + ": " + errorThrown);
			}
		});
	}

	// Private method to send a trickle candidate
	function sendTrickleCandidate(handleId, candidate) {
		if (!connected) {
			Janus.warn("Is the server down? (connected=false)");
			return;
		}
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			return;
		}
		var request = { "janus": "trickle", "candidate": candidate, "transaction": Janus.randomString(12) };
		if (pluginHandle.token) request["token"] = pluginHandle.token;
		if (apisecret) request["apisecret"] = apisecret;
		Janus.vdebug("Sending trickle candidate (handle=" + handleId + "):");
		Janus.vdebug(request);
		if (websockets) {
			request["session_id"] = sessionId;
			request["handle_id"] = handleId;
			ws.send(JSON.stringify(request));
			return;
		}
		Janus.httpAPICall(server + "/" + sessionId + "/" + handleId, {
			verb: 'POST',
			withCredentials: withCredentials,
			body: request,
			success: function success(json) {
				Janus.vdebug("Candidate sent!");
				Janus.vdebug(json);
				if (json["janus"] !== "ack") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
					return;
				}
			},
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown); // FIXME
			}
		});
	}

	// Private method to create a data channel
	function createDataChannel(handleId, dclabel, dcprotocol, incoming, pendingData) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			return;
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.pc) {
			Janus.warn("Invalid PeerConnection");
			return;
		}
		var onDataChannelMessage = function onDataChannelMessage(event) {
			Janus.log('Received message on data channel:', event);
			var label = event.target.label;
			pluginHandle.ondata(event.data, label);
		};
		var onDataChannelStateChange = function onDataChannelStateChange(event) {
			Janus.log('Received state change on data channel:', event);
			var label = event.target.label;
			var protocol = event.target.protocol;
			var dcState = config.dataChannel[label] ? config.dataChannel[label].readyState : "null";
			Janus.log('State change on <' + label + '> data channel: ' + dcState);
			if (dcState === 'open') {
				// Any pending messages to send?
				if (config.dataChannel[label].pending && config.dataChannel[label].pending.length > 0) {
					Janus.log("Sending pending messages on <" + label + ">:", config.dataChannel[label].pending.length);
					var _iteratorNormalCompletion4 = true;
					var _didIteratorError4 = false;
					var _iteratorError4 = undefined;

					try {
						for (var _iterator4 = config.dataChannel[label].pending[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
							var data = _step4.value;

							Janus.log("Sending data on data channel <" + label + ">");
							Janus.debug(data);
							config.dataChannel[label].send(data);
						}
					} catch (err) {
						_didIteratorError4 = true;
						_iteratorError4 = err;
					} finally {
						try {
							if (!_iteratorNormalCompletion4 && _iterator4["return"]) {
								_iterator4["return"]();
							}
						} finally {
							if (_didIteratorError4) {
								throw _iteratorError4;
							}
						}
					}

					config.dataChannel[label].pending = [];
				}
				// Notify the open data channel
				pluginHandle.ondataopen(label, protocol);
			}
		};
		var onDataChannelError = function onDataChannelError(error) {
			Janus.error('Got error on data channel:', error);
			// TODO
		};
		if (!incoming) {
			// FIXME Add options (ordered, maxRetransmits, etc.)
			var dcoptions = config.dataChannelOptions;
			if (dcprotocol) dcoptions.protocol = dcprotocol;
			config.dataChannel[dclabel] = config.pc.createDataChannel(dclabel, dcoptions);
		} else {
			// The channel was created by Janus
			config.dataChannel[dclabel] = incoming;
		}
		config.dataChannel[dclabel].onmessage = onDataChannelMessage;
		config.dataChannel[dclabel].onopen = onDataChannelStateChange;
		config.dataChannel[dclabel].onclose = onDataChannelStateChange;
		config.dataChannel[dclabel].onerror = onDataChannelError;
		config.dataChannel[dclabel].pending = [];
		if (pendingData) config.dataChannel[dclabel].pending.push(pendingData);
	}

	// Private method to send a data channel message
	function sendData(handleId, callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			callbacks.error("Invalid handle");
			return;
		}
		var config = pluginHandle.webrtcStuff;
		var data = callbacks.text || callbacks.data;
		if (!data) {
			Janus.warn("Invalid data");
			callbacks.error("Invalid data");
			return;
		}
		var label = callbacks.label ? callbacks.label : Janus.dataChanDefaultLabel;
		if (!config.dataChannel[label]) {
			// Create new data channel and wait for it to open
			createDataChannel(handleId, label, callbacks.protocol, false, data, callbacks.protocol);
			callbacks.success();
			return;
		}
		if (config.dataChannel[label].readyState !== "open") {
			config.dataChannel[label].pending.push(data);
			callbacks.success();
			return;
		}
		Janus.log("Sending data on data channel <" + label + ">");
		Janus.debug(data);
		config.dataChannel[label].send(data);
		callbacks.success();
	}

	// Private method to send a DTMF tone
	function sendDtmf(handleId, callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			callbacks.error("Invalid handle");
			return;
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.dtmfSender) {
			// Create the DTMF sender the proper way, if possible
			if (config.pc) {
				var senders = config.pc.getSenders();
				var audioSender = senders.find(function (sender) {
					return sender.track && sender.track.kind === 'audio';
				});
				if (!audioSender) {
					Janus.warn("Invalid DTMF configuration (no audio track)");
					callbacks.error("Invalid DTMF configuration (no audio track)");
					return;
				}
				config.dtmfSender = audioSender.dtmf;
				if (config.dtmfSender) {
					Janus.log("Created DTMF Sender");
					config.dtmfSender.ontonechange = function (tone) {
						Janus.debug("Sent DTMF tone: " + tone.tone);
					};
				}
			}
			if (!config.dtmfSender) {
				Janus.warn("Invalid DTMF configuration");
				callbacks.error("Invalid DTMF configuration");
				return;
			}
		}
		var dtmf = callbacks.dtmf;
		if (!dtmf) {
			Janus.warn("Invalid DTMF parameters");
			callbacks.error("Invalid DTMF parameters");
			return;
		}
		var tones = dtmf.tones;
		if (!tones) {
			Janus.warn("Invalid DTMF string");
			callbacks.error("Invalid DTMF string");
			return;
		}
		var duration = typeof dtmf.duration === 'number' ? dtmf.duration : 500; // We choose 500ms as the default duration for a tone
		var gap = typeof dtmf.gap === 'number' ? dtmf.gap : 50; // We choose 50ms as the default gap between tones
		Janus.debug("Sending DTMF string " + tones + " (duration " + duration + "ms, gap " + gap + "ms)");
		config.dtmfSender.insertDTMF(tones, duration, gap);
		callbacks.success();
	}

	// Private method to destroy a plugin handle
	function destroyHandle(handleId, callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		var noRequest = callbacks.noRequest === true;
		Janus.log("Destroying handle " + handleId + " (only-locally=" + noRequest + ")");
		cleanupWebrtc(handleId);
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || pluginHandle.detached) {
			// Plugin was already detached by Janus, calling detach again will return a handle not found error, so just exit here
			delete pluginHandles[handleId];
			callbacks.success();
			return;
		}
		pluginHandle.detached = true;
		if (noRequest) {
			// We're only removing the handle locally
			delete pluginHandles[handleId];
			callbacks.success();
			return;
		}
		if (!connected) {
			Janus.warn("Is the server down? (connected=false)");
			callbacks.error("Is the server down? (connected=false)");
			return;
		}
		var request = { "janus": "detach", "transaction": Janus.randomString(12) };
		if (pluginHandle.token) request["token"] = pluginHandle.token;
		if (apisecret) request["apisecret"] = apisecret;
		if (websockets) {
			request["session_id"] = sessionId;
			request["handle_id"] = handleId;
			ws.send(JSON.stringify(request));
			delete pluginHandles[handleId];
			callbacks.success();
			return;
		}
		Janus.httpAPICall(server + "/" + sessionId + "/" + handleId, {
			verb: 'POST',
			withCredentials: withCredentials,
			body: request,
			success: function success(json) {
				Janus.log("Destroyed handle:");
				Janus.debug(json);
				if (json["janus"] !== "success") {
					Janus.error("Ooops: " + json["error"].code + " " + json["error"].reason); // FIXME
				}
				delete pluginHandles[handleId];
				callbacks.success();
			},
			error: function error(textStatus, errorThrown) {
				Janus.error(textStatus + ":", errorThrown); // FIXME
				// We cleanup anyway
				delete pluginHandles[handleId];
				callbacks.success();
			}
		});
	}

	// WebRTC stuff
	// Helper function to create a new PeerConnection, if we need one
	function createPeerconnectionIfNeeded(handleId, callbacks) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			throw "Invalid handle";
		}
		var config = pluginHandle.webrtcStuff;
		if (config.pc) {
			// Nothing to do, we have a PeerConnection already
			return;
		}
		var pc_config = {
			iceServers: iceServers,
			iceTransportPolicy: iceTransportPolicy,
			bundlePolicy: bundlePolicy
		};
		pc_config.sdpSemantics = 'unified-plan';
		// Check if a sender or receiver transform has been provided
		var insertableStreams = false;
		if (callbacks.tracks) {
			var _iteratorNormalCompletion5 = true;
			var _didIteratorError5 = false;
			var _iteratorError5 = undefined;

			try {
				for (var _iterator5 = callbacks.tracks[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
					var track = _step5.value;

					if (track.transforms && (track.transforms.sender || track.transforms.receiver)) {
						insertableStreams = true;
						break;
					}
				}
			} catch (err) {
				_didIteratorError5 = true;
				_iteratorError5 = err;
			} finally {
				try {
					if (!_iteratorNormalCompletion5 && _iterator5["return"]) {
						_iterator5["return"]();
					}
				} finally {
					if (_didIteratorError5) {
						throw _iteratorError5;
					}
				}
			}
		}
		if (RTCRtpSender && (RTCRtpSender.prototype.createEncodedStreams || RTCRtpSender.prototype.createEncodedAudioStreams && RTCRtpSender.prototype.createEncodedVideoStreams) && insertableStreams) {
			config.insertableStreams = true;
			pc_config.forceEncodedAudioInsertableStreams = true;
			pc_config.forceEncodedVideoInsertableStreams = true;
			pc_config.encodedInsertableStreams = true;
		}
		Janus.log('Creating PeerConnection');
		config.pc = new RTCPeerConnection(pc_config);
		Janus.debug(config.pc);
		if (config.pc.getStats) {
			// FIXME
			config.volume = {};
			config.bitrate.value = '0 kbits/sec';
		}
		Janus.log('Preparing local SDP and gathering candidates (trickle=' + config.trickle + ')');
		config.pc.oniceconnectionstatechange = function () {
			if (config.pc) pluginHandle.iceState(config.pc.iceConnectionState);
		};
		config.pc.onicecandidate = function (event) {
			if (!event.candidate || event.candidate.candidate && event.candidate.candidate.indexOf('endOfCandidates') > 0) {
				Janus.log('End of candidates.');
				config.iceDone = true;
				if (config.trickle === true) {
					// Notify end of candidates
					sendTrickleCandidate(handleId, { completed: true });
				} else {
					// No trickle, time to send the complete SDP (including all candidates)
					sendSDP(handleId, callbacks);
				}
			} else {
				// JSON.stringify doesn't work on some WebRTC objects anymore
				// See https://code.google.com/p/chromium/issues/detail?id=467366
				var candidate = {
					candidate: event.candidate.candidate,
					sdpMid: event.candidate.sdpMid,
					sdpMLineIndex: event.candidate.sdpMLineIndex
				};
				if (config.trickle === true) {
					// Send candidate
					sendTrickleCandidate(handleId, candidate);
				}
			}
		};
		config.pc.ontrack = function (event) {
			Janus.log('Handling Remote Track', event);
			if (!event.streams) return;
			if (!event.track) return;
			// Notify about the new track event
			var mid = event.transceiver ? event.transceiver.mid : event.track.id;
			try {
				pluginHandle.onremotetrack(event.track, mid, true, { reason: 'created' });
			} catch (e) {
				Janus.error("Error calling onremotetrack", e);
			}
			if (event.track.onended) return;
			var trackMutedTimeoutId = null;
			Janus.log('Adding onended callback to track:', event.track);
			event.track.onended = function (ev) {
				Janus.log('Remote track removed:', ev);
				clearTimeout(trackMutedTimeoutId);
				// Notify the application
				var transceivers = config.pc ? config.pc.getTransceivers() : null;
				var transceiver = transceivers ? transceivers.find(function (t) {
					return t.receiver.track === ev.target;
				}) : null;
				var mid = transceiver ? transceiver.mid : ev.target.id;
				try {
					pluginHandle.onremotetrack(ev.target, mid, false, { reason: 'ended' });
				} catch (e) {
					Janus.error("Error calling onremotetrack on removal", e);
				}
			};
			event.track.onmute = function (ev) {
				Janus.log('Remote track muted:', ev);
				if (!trackMutedTimeoutId) {
					trackMutedTimeoutId = setTimeout(function () {
						Janus.log('Removing remote track');
						// Notify the application the track is gone
						var transceivers = config.pc ? config.pc.getTransceivers() : null;
						var transceiver = transceivers ? transceivers.find(function (t) {
							return t.receiver.track === ev.target;
						}) : null;
						var mid = transceiver ? transceiver.mid : ev.target.id;
						try {
							pluginHandle.onremotetrack(ev.target, mid, false, { reason: 'mute' });
						} catch (e) {
							Janus.error("Error calling onremotetrack on mute", e);
						}
						trackMutedTimeoutId = null;
						// Chrome seems to raise mute events only at multiples of 834ms;
						// we set the timeout to three times this value (rounded to 840ms)
					}, 3 * 840);
				}
			};
			event.track.onunmute = function (ev) {
				Janus.log('Remote track flowing again:', ev);
				if (trackMutedTimeoutId != null) {
					clearTimeout(trackMutedTimeoutId);
					trackMutedTimeoutId = null;
				} else {
					try {
						// Notify the application the track is back
						var transceivers = config.pc ? config.pc.getTransceivers() : null;
						var transceiver = transceivers ? transceivers.find(function (t) {
							return t.receiver.track === ev.target;
						}) : null;
						var _mid = transceiver ? transceiver.mid : ev.target.id;
						pluginHandle.onremotetrack(ev.target, _mid, true, { reason: 'unmute' });
					} catch (e) {
						Janus.error("Error calling onremotetrack on unmute", e);
					}
				}
			};
		};
	}

	// Helper function used when creating either an offer or answer: it
	// prepares what needs to be prepared, including creating a new
	// PeerConnection (if needed) and updating the tracks configuration,
	// before invoking the function to actually generate the offer/answer
	function prepareWebrtc(handleId, offer, callbacks) {
		var jsep, _iteratorNormalCompletion6, _didIteratorError6, _iteratorError6, _iterator6, _step6, track, pluginHandle, config, _offer, i, candidate, answer;

		return regeneratorRuntime.async(function prepareWebrtc$(context$2$0) {
			while (1) switch (context$2$0.prev = context$2$0.next) {
				case 0:
					callbacks = callbacks || {};
					callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
					callbacks.error = typeof callbacks.error == "function" ? callbacks.error : webrtcError;
					jsep = callbacks.jsep;

					if (!(offer && jsep)) {
						context$2$0.next = 10;
						break;
					}

					Janus.error("Provided a JSEP to a createOffer");
					callbacks.error("Provided a JSEP to a createOffer");
					return context$2$0.abrupt("return");

				case 10:
					if (!(!offer && (!jsep || !jsep.type || !jsep.sdp))) {
						context$2$0.next = 14;
						break;
					}

					Janus.error("A valid JSEP is required for createAnswer");
					callbacks.error("A valid JSEP is required for createAnswer");
					return context$2$0.abrupt("return");

				case 14:
					if (!(callbacks.media && !callbacks.tracks)) {
						context$2$0.next = 45;
						break;
					}

					callbacks.tracks = Janus.mediaToTracks(callbacks.media);

					if (!(callbacks.simulcast === true || callbacks.simulcast2 === true || callbacks.svc)) {
						context$2$0.next = 44;
						break;
					}

					_iteratorNormalCompletion6 = true;
					_didIteratorError6 = false;
					_iteratorError6 = undefined;
					context$2$0.prev = 20;
					_iterator6 = callbacks.tracks[Symbol.iterator]();

				case 22:
					if (_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done) {
						context$2$0.next = 30;
						break;
					}

					track = _step6.value;

					if (!(track.type === 'video')) {
						context$2$0.next = 27;
						break;
					}

					if (callbacks.simulcast === true || callbacks.simulcast2 === true) track.simulcast = true;else if (callbacks.svc) track.svc = callbacks.svc;
					return context$2$0.abrupt("break", 30);

				case 27:
					_iteratorNormalCompletion6 = true;
					context$2$0.next = 22;
					break;

				case 30:
					context$2$0.next = 36;
					break;

				case 32:
					context$2$0.prev = 32;
					context$2$0.t0 = context$2$0["catch"](20);
					_didIteratorError6 = true;
					_iteratorError6 = context$2$0.t0;

				case 36:
					context$2$0.prev = 36;
					context$2$0.prev = 37;

					if (!_iteratorNormalCompletion6 && _iterator6["return"]) {
						_iterator6["return"]();
					}

				case 39:
					context$2$0.prev = 39;

					if (!_didIteratorError6) {
						context$2$0.next = 42;
						break;
					}

					throw _iteratorError6;

				case 42:
					return context$2$0.finish(39);

				case 43:
					return context$2$0.finish(36);

				case 44:
					Janus.warn('Deprecated media object passed, use tracks instead. Automatically translated to:', callbacks.tracks);

				case 45:
					if (!(callbacks.tracks && !Array.isArray(callbacks.tracks))) {
						context$2$0.next = 49;
						break;
					}

					Janus.error("Tracks must be an array");
					callbacks.error("Tracks must be an array");
					return context$2$0.abrupt("return");

				case 49:
					pluginHandle = pluginHandles[handleId];

					if (!(!pluginHandle || !pluginHandle.webrtcStuff)) {
						context$2$0.next = 54;
						break;
					}

					Janus.warn("Invalid handle");
					callbacks.error("Invalid handle");
					return context$2$0.abrupt("return");

				case 54:
					config = pluginHandle.webrtcStuff;

					config.trickle = isTrickleEnabled(callbacks.trickle);
					context$2$0.prev = 56;

					// Create a PeerConnection, if needed
					createPeerconnectionIfNeeded(handleId, callbacks);

					if (!offer) {
						context$2$0.next = 61;
						break;
					}

					context$2$0.next = 61;
					return regeneratorRuntime.awrap(captureDevices(handleId, callbacks));

				case 61:
					if (jsep) {
						context$2$0.next = 68;
						break;
					}

					context$2$0.next = 64;
					return regeneratorRuntime.awrap(createOffer(handleId, callbacks));

				case 64:
					_offer = context$2$0.sent;

					callbacks.success(_offer);
					context$2$0.next = 79;
					break;

				case 68:
					context$2$0.next = 70;
					return regeneratorRuntime.awrap(config.pc.setRemoteDescription(jsep));

				case 70:
					Janus.log("Remote description accepted!");
					config.remoteSdp = jsep.sdp;
					// Any trickle candidate we cached?
					if (config.candidates && config.candidates.length > 0) {
						for (i = 0; i < config.candidates.length; i++) {
							candidate = config.candidates[i];

							Janus.debug("Adding remote candidate:", candidate);
							if (!candidate || candidate.completed === true) {
								// end-of-candidates
								config.pc.addIceCandidate(Janus.endOfCandidates);
							} else {
								// New candidate
								config.pc.addIceCandidate(candidate);
							}
						}
						config.candidates = [];
					}
					// Capture devices and setup tracks, if needed
					context$2$0.next = 75;
					return regeneratorRuntime.awrap(captureDevices(handleId, callbacks));

				case 75:
					context$2$0.next = 77;
					return regeneratorRuntime.awrap(createAnswer(handleId, callbacks));

				case 77:
					answer = context$2$0.sent;

					callbacks.success(answer);

				case 79:
					context$2$0.next = 85;
					break;

				case 81:
					context$2$0.prev = 81;
					context$2$0.t1 = context$2$0["catch"](56);

					Janus.error(context$2$0.t1);
					callbacks.error(context$2$0.t1);

				case 85:
				case "end":
					return context$2$0.stop();
			}
		}, null, this, [[20, 32, 36, 44], [37,, 39, 43], [56, 81]]);
	}

	function prepareWebrtcPeer(handleId, callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : webrtcError;
		callbacks.customizeSdp = typeof callbacks.customizeSdp == "function" ? callbacks.customizeSdp : Janus.noop;
		var jsep = callbacks.jsep;
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			callbacks.error("Invalid handle");
			return;
		}
		var config = pluginHandle.webrtcStuff;
		if (jsep) {
			if (!config.pc) {
				Janus.warn("Wait, no PeerConnection?? if this is an answer, use createAnswer and not handleRemoteJsep");
				callbacks.error("No PeerConnection: if this is an answer, use createAnswer and not handleRemoteJsep");
				return;
			}

			// MY_CODE: MUNGING SDP
			jsep.sdp = (0, _sdpHelper.setMaxAudioBitrate)(jsep.sdp);

			callbacks.customizeSdp(jsep);
			config.pc.setRemoteDescription(jsep).then(function () {
				Janus.log("Remote description accepted!");
				config.remoteSdp = jsep.sdp;
				// Any trickle candidate we cached?
				if (config.candidates && config.candidates.length > 0) {
					for (var i = 0; i < config.candidates.length; i++) {
						var candidate = config.candidates[i];
						Janus.debug("Adding remote candidate:", candidate);
						if (!candidate || candidate.completed === true) {
							// end-of-candidates
							config.pc.addIceCandidate(Janus.endOfCandidates);
						} else {
							// New candidate
							config.pc.addIceCandidate(candidate);
						}
					}
					config.candidates = [];
				}
				// Done
				callbacks.success();
			}, callbacks.error);
		} else {
			callbacks.error("Invalid JSEP");
		}
	}

	function createOffer(handleId, callbacks) {
		var pluginHandle, config, mediaConstraints, iceRestart, offer, jsep;
		return regeneratorRuntime.async(function createOffer$(context$2$0) {
			while (1) switch (context$2$0.prev = context$2$0.next) {
				case 0:
					callbacks = callbacks || {};
					callbacks.customizeSdp = typeof callbacks.customizeSdp == "function" ? callbacks.customizeSdp : Janus.noop;
					pluginHandle = pluginHandles[handleId];

					if (!(!pluginHandle || !pluginHandle.webrtcStuff)) {
						context$2$0.next = 6;
						break;
					}

					Janus.warn("Invalid handle");
					throw "Invalid handle";

				case 6:
					config = pluginHandle.webrtcStuff;

					Janus.log("Creating offer (iceDone=" + config.iceDone + ")");
					// https://code.google.com/p/webrtc/issues/detail?id=3508
					mediaConstraints = {};
					iceRestart = callbacks.iceRestart === true;

					// If we need an ICE restart, set the related constraint
					if (iceRestart) mediaConstraints.iceRestart = true;
					Janus.debug(mediaConstraints);
					context$2$0.next = 14;
					return regeneratorRuntime.awrap(config.pc.createOffer(mediaConstraints));

				case 14:
					offer = context$2$0.sent;

					Janus.debug(offer);
					// JSON.stringify doesn't work on some WebRTC objects anymore
					// See https://code.google.com/p/chromium/issues/detail?id=467366
					jsep = {
						type: 'offer',
						sdp: offer.sdp
					};

					callbacks.customizeSdp(jsep);
					offer.sdp = jsep.sdp;
					Janus.log("Setting local description");

					// MY_CODE: MUNGING SDP
					jsep.sdp = (0, _sdpHelper.setMaxAudioBitrate)(jsep.sdp);

					config.mySdp = {
						type: 'offer',
						sdp: offer.sdp
					};
					context$2$0.next = 24;
					return regeneratorRuntime.awrap(config.pc.setLocalDescription(offer));

				case 24:
					config.mediaConstraints = mediaConstraints;

					if (!(!config.iceDone && !config.trickle)) {
						context$2$0.next = 28;
						break;
					}

					// FIXME Don't do anything until we have all candidates
					Janus.log("Waiting for all candidates...");
					return context$2$0.abrupt("return", null);

				case 28:
					// If transforms are present, notify Janus that the media is end-to-end encrypted
					if (config.insertableStreams) offer.e2ee = true;
					return context$2$0.abrupt("return", offer);

				case 30:
				case "end":
					return context$2$0.stop();
			}
		}, null, this);
	}

	function createAnswer(handleId, callbacks) {
		var pluginHandle, config, answer, jsep;
		return regeneratorRuntime.async(function createAnswer$(context$2$0) {
			while (1) switch (context$2$0.prev = context$2$0.next) {
				case 0:
					callbacks = callbacks || {};
					callbacks.customizeSdp = typeof callbacks.customizeSdp == "function" ? callbacks.customizeSdp : Janus.noop;
					pluginHandle = pluginHandles[handleId];

					if (!(!pluginHandle || !pluginHandle.webrtcStuff)) {
						context$2$0.next = 6;
						break;
					}

					Janus.warn("Invalid handle");
					throw "Invalid handle";

				case 6:
					config = pluginHandle.webrtcStuff;

					Janus.log("Creating answer (iceDone=" + config.iceDone + ")");
					context$2$0.next = 10;
					return regeneratorRuntime.awrap(config.pc.createAnswer());

				case 10:
					answer = context$2$0.sent;

					Janus.debug(answer);
					// JSON.stringify doesn't work on some WebRTC objects anymore
					// See https://code.google.com/p/chromium/issues/detail?id=467366
					jsep = {
						type: 'answer',
						sdp: answer.sdp
					};

					callbacks.customizeSdp(jsep);
					answer.sdp = jsep.sdp;
					Janus.log("Setting local description");
					config.mySdp = {
						type: 'answer',
						sdp: answer.sdp
					};
					context$2$0.next = 19;
					return regeneratorRuntime.awrap(config.pc.setLocalDescription(answer));

				case 19:
					if (!(!config.iceDone && !config.trickle)) {
						context$2$0.next = 22;
						break;
					}

					// FIXME Don't do anything until we have all candidates
					Janus.log("Waiting for all candidates...");
					return context$2$0.abrupt("return", null);

				case 22:
					// If transforms are present, notify Janus that the media is end-to-end encrypted
					if (config.insertableStreams) answer.e2ee = true;
					return context$2$0.abrupt("return", answer);

				case 24:
				case "end":
					return context$2$0.stop();
			}
		}, null, this);
	}

	function sendSDP(handleId, callbacks) {
		callbacks = callbacks || {};
		callbacks.success = typeof callbacks.success == "function" ? callbacks.success : Janus.noop;
		callbacks.error = typeof callbacks.error == "function" ? callbacks.error : Janus.noop;
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle, not sending anything");
			return;
		}
		var config = pluginHandle.webrtcStuff;
		Janus.log("Sending offer/answer SDP...");
		if (!config.mySdp) {
			Janus.warn("Local SDP instance is invalid, not sending anything...");
			return;
		}
		config.mySdp = {
			type: config.pc.localDescription.type,
			sdp: config.pc.localDescription.sdp
		};
		if (config.trickle === false) config.mySdp["trickle"] = false;
		Janus.debug(callbacks);
		config.sdpSent = true;
		callbacks.success(config.mySdp);
	}

	function _replaceTracks(handleId, callbacks) {
		var _iteratorNormalCompletion7, _didIteratorError7, _iteratorError7, _iterator7, _step7, track;

		return regeneratorRuntime.async(function _replaceTracks$(context$2$0) {
			while (1) switch (context$2$0.prev = context$2$0.next) {
				case 0:
					callbacks = callbacks || {};
					callbacks.success = typeof callbacks.success == 'function' ? callbacks.success : Janus.noop;
					callbacks.error = typeof callbacks.error == 'function' ? callbacks.error : Janus.noop;
					// Check that callbacks.array is a valid array

					if (!(callbacks.tracks && !Array.isArray(callbacks.tracks))) {
						context$2$0.next = 7;
						break;
					}

					Janus.error('Tracks must be an array');
					callbacks.error('Tracks must be an array');
					return context$2$0.abrupt("return");

				case 7:
					_iteratorNormalCompletion7 = true;
					_didIteratorError7 = false;
					_iteratorError7 = undefined;
					context$2$0.prev = 10;

					// Add the replace:true if it's missing
					for (_iterator7 = callbacks.tracks[Symbol.iterator](); !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
						track = _step7.value;

						if (track.add || !track.replace && !track.remove) track.replace = true;
					}
					context$2$0.next = 18;
					break;

				case 14:
					context$2$0.prev = 14;
					context$2$0.t0 = context$2$0["catch"](10);
					_didIteratorError7 = true;
					_iteratorError7 = context$2$0.t0;

				case 18:
					context$2$0.prev = 18;
					context$2$0.prev = 19;

					if (!_iteratorNormalCompletion7 && _iterator7["return"]) {
						_iterator7["return"]();
					}

				case 21:
					context$2$0.prev = 21;

					if (!_didIteratorError7) {
						context$2$0.next = 24;
						break;
					}

					throw _iteratorError7;

				case 24:
					return context$2$0.finish(21);

				case 25:
					return context$2$0.finish(18);

				case 26:
					context$2$0.prev = 26;
					context$2$0.next = 29;
					return regeneratorRuntime.awrap(captureDevices(handleId, callbacks));

				case 29:
					callbacks.success();
					context$2$0.next = 36;
					break;

				case 32:
					context$2$0.prev = 32;
					context$2$0.t1 = context$2$0["catch"](26);

					Janus.error(context$2$0.t1);
					callbacks.error(context$2$0.t1);

				case 36:
				case "end":
					return context$2$0.stop();
			}
		}, null, this, [[10, 14, 18, 26], [19,, 21, 25], [26, 32]]);
	}

	function captureDevices(handleId, callbacks) {
		var pluginHandle, config, tracks, openedConsentDialog, groups, _iteratorNormalCompletion8, _didIteratorError8, _iteratorError8, _iterator8, _step8, track, group, keys, _iteratorNormalCompletion9, _didIteratorError9, _iteratorError9, _iterator9, _step9, key, answer, _iteratorNormalCompletion10, _didIteratorError10, _iteratorError10, _loop, _iterator10, _step10, _iteratorNormalCompletion11, _didIteratorError11, _iteratorError11, _iterator11, _step11, _iteratorNormalCompletion12, _didIteratorError12, _iteratorError12, _iterator12, _step12, _ret4;

		return regeneratorRuntime.async(function captureDevices$(context$2$0) {
			var _this = this;

			while (1) switch (context$2$0.prev = context$2$0.next) {
				case 0:
					pluginHandle = pluginHandles[handleId];

					if (!(!pluginHandle || !pluginHandle.webrtcStuff)) {
						context$2$0.next = 4;
						break;
					}

					Janus.warn('Invalid handle, not sending anything');
					throw 'Invalid handle';

				case 4:
					config = pluginHandle.webrtcStuff;

					if (config.pc) {
						context$2$0.next = 8;
						break;
					}

					Janus.warn('Invalid PeerConnection');
					throw 'Invalid PeerConnection';

				case 8:
					tracks = callbacks.tracks;

					if (!(!tracks || !Array.isArray(tracks) || tracks.length === 0)) {
						context$2$0.next = 11;
						break;
					}

					return context$2$0.abrupt("return");

				case 11:
					openedConsentDialog = false;
					groups = {};
					_iteratorNormalCompletion8 = true;
					_didIteratorError8 = false;
					_iteratorError8 = undefined;
					context$2$0.prev = 16;
					_iterator8 = tracks[Symbol.iterator]();

				case 18:
					if (_iteratorNormalCompletion8 = (_step8 = _iterator8.next()).done) {
						context$2$0.next = 34;
						break;
					}

					track = _step8.value;

					delete track.gumGroup;

					if (!(!track.type || !['audio', 'video'].includes(track.type))) {
						context$2$0.next = 23;
						break;
					}

					return context$2$0.abrupt("continue", 31);

				case 23:
					if (!(!track.capture || track.capture instanceof MediaStreamTrack)) {
						context$2$0.next = 25;
						break;
					}

					return context$2$0.abrupt("continue", 31);

				case 25:
					group = track.group ? track.group : 'default';

					if (!groups[group]) groups[group] = {};

					if (!groups[group][track.type]) {
						context$2$0.next = 29;
						break;
					}

					return context$2$0.abrupt("continue", 31);

				case 29:
					track.gumGroup = group;
					groups[group][track.type] = track;

				case 31:
					_iteratorNormalCompletion8 = true;
					context$2$0.next = 18;
					break;

				case 34:
					context$2$0.next = 40;
					break;

				case 36:
					context$2$0.prev = 36;
					context$2$0.t0 = context$2$0["catch"](16);
					_didIteratorError8 = true;
					_iteratorError8 = context$2$0.t0;

				case 40:
					context$2$0.prev = 40;
					context$2$0.prev = 41;

					if (!_iteratorNormalCompletion8 && _iterator8["return"]) {
						_iterator8["return"]();
					}

				case 43:
					context$2$0.prev = 43;

					if (!_didIteratorError8) {
						context$2$0.next = 46;
						break;
					}

					throw _iteratorError8;

				case 46:
					return context$2$0.finish(43);

				case 47:
					return context$2$0.finish(40);

				case 48:
					keys = Object.keys(groups);
					_iteratorNormalCompletion9 = true;
					_didIteratorError9 = false;
					_iteratorError9 = undefined;
					context$2$0.prev = 52;

					for (_iterator9 = keys[Symbol.iterator](); !(_iteratorNormalCompletion9 = (_step9 = _iterator9.next()).done); _iteratorNormalCompletion9 = true) {
						key = _step9.value;
						group = groups[key];

						if (!group.audio || !group.video) {
							if (group.audio) delete group.audio.gumGroup;
							if (group.video) delete group.video.gumGroup;
							delete groups[key];
						}
					}
					context$2$0.next = 60;
					break;

				case 56:
					context$2$0.prev = 56;
					context$2$0.t1 = context$2$0["catch"](52);
					_didIteratorError9 = true;
					_iteratorError9 = context$2$0.t1;

				case 60:
					context$2$0.prev = 60;
					context$2$0.prev = 61;

					if (!_iteratorNormalCompletion9 && _iterator9["return"]) {
						_iterator9["return"]();
					}

				case 63:
					context$2$0.prev = 63;

					if (!_didIteratorError9) {
						context$2$0.next = 66;
						break;
					}

					throw _iteratorError9;

				case 66:
					return context$2$0.finish(63);

				case 67:
					return context$2$0.finish(60);

				case 68:
					answer = callbacks.jsep ? true : false;
					_iteratorNormalCompletion10 = true;
					_didIteratorError10 = false;
					_iteratorError10 = undefined;
					context$2$0.prev = 72;

					_loop = function callee$2$0() {
						var track, kind, transceiver, sender, nt, trackId, stream, constraints, otherType, otherTrack, otherConstraints, newDirection, maxBitrates, parameters, params, senderStreams, receiverStreams, rt, t, curdir, newdir, send, recv;
						return regeneratorRuntime.async(function callee$2$0$(context$3$0) {
							while (1) switch (context$3$0.prev = context$3$0.next) {
								case 0:
									track = _step10.value;

									if (track.type) {
										context$3$0.next = 4;
										break;
									}

									Janus.warn('Missing track type:', track);
									return context$3$0.abrupt("return", "continue");

								case 4:
									if (!(track.type === 'data')) {
										context$3$0.next = 12;
										break;
									}

									if (!config.pc.ondatachannel) {
										context$3$0.next = 8;
										break;
									}

									Janus.warn('Data channel exists already, not creating another one');
									return context$3$0.abrupt("return", "continue");

								case 8:
									Janus.log('Creating default data channel');
									createDataChannel(handleId, Janus.dataChanDefaultLabel, null, false);
									config.pc.ondatachannel = function (event) {
										Janus.log('Data channel created by Janus:', event);
										createDataChannel(handleId, event.channel.label, event.channel.protocol, event.channel);
									};
									return context$3$0.abrupt("return", "continue");

								case 12:
									if (typeof track.add === 'undefined' && track.add === null && typeof track.remove === 'undefined' && track.remove === null && typeof track.replace === 'undefined' && track.replace === null) {
										// Let's default to 'add'
										track.add = true;
									}

									if (!(track.add && track.remove || track.add && track.remove && track.replace)) {
										context$3$0.next = 16;
										break;
									}

									Janus.warn('Conflicting actions for track, ignoring:', track);
									return context$3$0.abrupt("return", "continue");

								case 16:
									if (track.add && track.replace) {
										Janus.warn('Both add and replace provided, falling back to replace:', track);
										delete track.add;
									} else if (track.remove && track.replace) {
										Janus.warn('Both remove and replace provided, falling back to remove:', track);
										delete track.replace;
									}
									kind = track.type;

									if (track.type === 'screen') kind = 'video'; // FIXME
									transceiver = null, sender = null;

									if (track.mid) {
										// Search by mid
										transceiver = config.pc.getTransceivers().find(function (t) {
											return t.mid === track.mid && t.receiver.track.kind === kind;
										});
									} else {
										// Find the first track of this type
										transceiver = config.pc.getTransceivers().find(function (t) {
											return t.receiver.track.kind === kind;
										});
									}

									if (!(track.replace || track.remove)) {
										context$3$0.next = 29;
										break;
									}

									if (transceiver) {
										context$3$0.next = 25;
										break;
									}

									Janus.warn("Couldn't find a transceiver for track:", track);
									return context$3$0.abrupt("return", "continue");

								case 25:
									if (transceiver.sender) {
										context$3$0.next = 28;
										break;
									}

									Janus.warn('No sender in the transceiver for track:', track);
									return context$3$0.abrupt("return", "continue");

								case 28:
									sender = transceiver.sender;

								case 29:
									if (!(answer && !transceiver)) {
										context$3$0.next = 34;
										break;
									}

									transceiver = config.pc.getTransceivers().find(function (t) {
										return t.receiver.track.kind === kind;
									});

									if (transceiver) {
										context$3$0.next = 34;
										break;
									}

									Janus.warn("Couldn't find a transceiver for track:", track);
									return context$3$0.abrupt("return", "continue");

								case 34:
									nt = null, trackId = null;

									if (!track.remove) {
										context$3$0.next = 42;
										break;
									}

									Janus.log('Removing track from PeerConnection', track);
									trackId = sender.track ? sender.track.id : null;
									context$3$0.next = 40;
									return regeneratorRuntime.awrap(sender.replaceTrack(null));

								case 40:
									context$3$0.next = 122;
									break;

								case 42:
									if (!track.capture) {
										context$3$0.next = 121;
										break;
									}

									if (!(track.gumGroup && groups[track.gumGroup] && groups[track.gumGroup].stream)) {
										context$3$0.next = 51;
										break;
									}

									stream = groups[track.gumGroup].stream;

									nt = track.type === 'audio' ? stream.getAudioTracks()[0] : stream.getVideoTracks()[0];
									delete groups[track.gumGroup].stream;
									delete groups[track.gumGroup];
									delete track.gumGroup;
									context$3$0.next = 69;
									break;

								case 51:
									if (!(track.capture instanceof MediaStreamTrack)) {
										context$3$0.next = 55;
										break;
									}

									// An external track was provided, use that
									nt = track.capture;
									context$3$0.next = 69;
									break;

								case 55:
									if (!openedConsentDialog) {
										openedConsentDialog = true;
										pluginHandle.consentDialog(true);
									}
									constraints = Janus.trackConstraints(track), stream = null;

									if (!(track.type === 'audio' || track.type === 'video')) {
										context$3$0.next = 65;
										break;
									}

									// Use getUserMedia: check if we need to group audio and video together
									if (track.gumGroup) {
										otherType = track.type === 'audio' ? 'video' : 'audio';

										if (groups[track.gumGroup] && groups[track.gumGroup][otherType]) {
											otherTrack = groups[track.gumGroup][otherType];
											otherConstraints = Janus.trackConstraints(otherTrack);

											constraints[otherType] = otherConstraints[otherType];
										}
									}
									context$3$0.next = 61;
									return regeneratorRuntime.awrap(navigator.mediaDevices.getUserMedia(constraints));

								case 61:
									stream = context$3$0.sent;

									if (track.gumGroup && constraints.audio && constraints.video) {
										// We just performed a grouped getUserMedia, keep track of the
										// stream so that we can immediately assign the track later
										groups[track.gumGroup].stream = stream;
										delete track.gumGroup;
									}
									context$3$0.next = 68;
									break;

								case 65:
									context$3$0.next = 67;
									return regeneratorRuntime.awrap(navigator.mediaDevices.getDisplayMedia(constraints));

								case 67:
									stream = context$3$0.sent;

								case 68:
									nt = track.type === 'audio' ? stream.getAudioTracks()[0] : stream.getVideoTracks()[0];

								case 69:
									if (!track.replace) {
										context$3$0.next = 77;
										break;
									}

									context$3$0.next = 72;
									return regeneratorRuntime.awrap(sender.replaceTrack(nt));

								case 72:
									newDirection = 'sendrecv';

									if (track.recv === false || transceiver.direction === 'inactive' || transceiver.direction === 'sendonly') newDirection = 'sendonly';
									if (transceiver.setDirection) transceiver.setDirection(newDirection);else transceiver.direction = newDirection;
									context$3$0.next = 118;
									break;

								case 77:
									// FIXME Add as a new track
									if (!config.myStream) config.myStream = new MediaStream();
									if (kind === 'audio' || !track.simulcast && !track.svc) {
										sender = config.pc.addTrack(nt, config.myStream);
										transceiver = config.pc.getTransceivers().find(function (t) {
											return t.sender === sender;
										});
									} else if (track.simulcast) {
										if (Janus.webRTCAdapter.browserDetails.browser !== 'firefox') {
											// Standard RID
											Janus.log('Enabling rid-based simulcasting:', nt);
											maxBitrates = getMaxBitrates(track.simulcastMaxBitrates);

											transceiver = config.pc.addTransceiver(nt, {
												direction: 'sendrecv',
												streams: [config.myStream],
												sendEncodings: track.sendEncodings || [{ rid: 'h', active: true, maxBitrate: maxBitrates.high }, { rid: 'm', active: true, maxBitrate: maxBitrates.medium, scaleResolutionDownBy: 2 }, { rid: 'l', active: true, maxBitrate: maxBitrates.low, scaleResolutionDownBy: 4 }]
											});
										} else {
											// Firefox-based RID, based on https://gist.github.com/voluntas/088bc3cc62094730647b
											Janus.log('Enabling Simulcasting for Firefox (RID)');
											transceiver = config.pc.addTransceiver(nt, {
												direction: 'sendrecv',
												streams: [config.myStream]
											});
											sender = transceiver ? transceiver.sender : null;
											if (sender) {
												parameters = sender.getParameters();

												if (!parameters) parameters = {};
												maxBitrates = getMaxBitrates(track.simulcastMaxBitrates);

												parameters.encodings = track.sendEncodings || [{ rid: 'h', active: true, maxBitrate: maxBitrates.high }, { rid: 'm', active: true, maxBitrate: maxBitrates.medium, scaleResolutionDownBy: 2 }, { rid: 'l', active: true, maxBitrate: maxBitrates.low, scaleResolutionDownBy: 4 }];
												sender.setParameters(parameters);
											}
										}
									} else {
										Janus.log('Enabling SVC (' + track.svc + '):', nt);
										transceiver = config.pc.addTransceiver(nt, {
											direction: 'sendrecv',
											streams: [config.myStream],
											sendEncodings: [{ scalabilityMode: track.svc }]
										});
									}
									if (!sender) sender = transceiver ? transceiver.sender : null;
									// Check if we need to override some settings
									if (track.codec) {
										if (Janus.webRTCAdapter.browserDetails.browser === 'firefox') {
											Janus.warn('setCodecPreferences not supported in Firefox, ignoring codec for track:', track);
										} else if (typeof track.codec !== 'string') {
											Janus.warn('Invalid codec value, ignoring for track:', track);
										} else {
											(function () {
												var mimeType = kind + '/' + track.codec.toLowerCase();
												var codecs = RTCRtpReceiver.getCapabilities(kind).codecs.filter(function (codec) {
													return codec.mimeType.toLowerCase() === mimeType;
												});
												if (!codecs || codecs.length === 0) {
													Janus.warn('Codec not supported in this browser for this track, ignoring:', track);
												} else if (transceiver) {
													try {
														transceiver.setCodecPreferences(codecs);
													} catch (err) {
														Janus.warn('Failed enforcing codec for this ' + kind + ' track:', err);
													}
												}
											})();
										}
									}

									if (!track.bitrate) {
										context$3$0.next = 99;
										break;
									}

									if (!(track.simulcast || track.svc)) {
										context$3$0.next = 86;
										break;
									}

									Janus.warn('Ignoring bitrate for simulcast/SVC track, use sendEncodings for that');
									context$3$0.next = 99;
									break;

								case 86:
									if (!(isNaN(track.bitrate) || track.bitrate < 0)) {
										context$3$0.next = 90;
										break;
									}

									Janus.warn('Ignoring invalid bitrate for track:', track);
									context$3$0.next = 99;
									break;

								case 90:
									if (!sender) {
										context$3$0.next = 99;
										break;
									}

									params = sender.getParameters();

									if (!(!params || !params.encodings || params.encodings.length === 0)) {
										context$3$0.next = 96;
										break;
									}

									Janus.warn('No encodings in the sender parameters, ignoring bitrate for track:', track);
									context$3$0.next = 99;
									break;

								case 96:
									params.encodings[0].maxBitrate = track.bitrate;
									context$3$0.next = 99;
									return regeneratorRuntime.awrap(sender.setParameters(params));

								case 99:
									if (!(kind === 'video' && track.framerate)) {
										context$3$0.next = 117;
										break;
									}

									if (!(track.simulcast || track.svc)) {
										context$3$0.next = 104;
										break;
									}

									Janus.warn('Ignoring framerate for simulcast/SVC track, use sendEncodings for that');
									context$3$0.next = 117;
									break;

								case 104:
									if (!(isNaN(track.framerate) || track.framerate < 0)) {
										context$3$0.next = 108;
										break;
									}

									Janus.warn('Ignoring invalid framerate for track:', track);
									context$3$0.next = 117;
									break;

								case 108:
									if (!sender) {
										context$3$0.next = 117;
										break;
									}

									params = sender.getParameters();

									if (!(!params || !params.encodings || params.encodings.length === 0)) {
										context$3$0.next = 114;
										break;
									}

									Janus.warn('No encodings in the sender parameters, ignoring framerate for track:', track);
									context$3$0.next = 117;
									break;

								case 114:
									params.encodings[0].maxFramerate = track.framerate;
									context$3$0.next = 117;
									return regeneratorRuntime.awrap(sender.setParameters(params));

								case 117:
									// Check if insertable streams are involved
									if (track.transforms) {
										if (sender && track.transforms.sender) {
											senderStreams = null;

											if (RTCRtpSender.prototype.createEncodedStreams) {
												senderStreams = sender.createEncodedStreams();
											} else if (RTCRtpSender.prototype.createAudioEncodedStreams || RTCRtpSender.prototype.createEncodedVideoStreams) {
												if (kind === 'audio') {
													senderStreams = sender.createEncodedAudioStreams();
												} else if (kind === 'video') {
													senderStreams = sender.createEncodedVideoStreams();
												}
											}
											if (senderStreams) {
												console.log('Insertable Streams sender transform:', senderStreams);
												if (senderStreams.readableStream && senderStreams.writableStream) {
													senderStreams.readableStream.pipeThrough(track.transforms.sender).pipeTo(senderStreams.writableStream);
												} else if (senderStreams.readable && senderStreams.writable) {
													senderStreams.readable.pipeThrough(track.transforms.sender).pipeTo(senderStreams.writable);
												}
											}
										}
										if (transceiver && transceiver.receiver && track.transforms.receiver) {
											receiverStreams = null;

											if (RTCRtpReceiver.prototype.createEncodedStreams) {
												receiverStreams = transceiver.receiver.createEncodedStreams();
											} else if (RTCRtpReceiver.prototype.createAudioEncodedStreams || RTCRtpReceiver.prototype.createEncodedVideoStreams) {
												if (kind === 'audio') {
													receiverStreams = transceiver.receiver.createEncodedAudioStreams();
												} else if (kind === 'video') {
													receiverStreams = transceiver.receiver.createEncodedVideoStreams();
												}
											}
											if (receiverStreams) {
												console.log('Insertable Streams receiver transform:', receiverStreams);
												if (receiverStreams.readableStream && receiverStreams.writableStream) {
													receiverStreams.readableStream.pipeThrough(track.transforms.receiver).pipeTo(receiverStreams.writableStream);
												} else if (receiverStreams.readable && receiverStreams.writable) {
													receiverStreams.readable.pipeThrough(track.transforms.receiver).pipeTo(receiverStreams.writable);
												}
											}
										}
									}

								case 118:
									if (nt && track.dontStop === true) nt.dontStop = true;
									context$3$0.next = 122;
									break;

								case 121:
									if (track.recv && !transceiver) {
										// Maybe a new recvonly track
										transceiver = config.pc.addTransceiver(kind);
										if (transceiver) {
											// Check if we need to override some settings
											if (track.codec) {
												if (Janus.webRTCAdapter.browserDetails.browser === 'firefox') {
													Janus.warn('setCodecPreferences not supported in Firefox, ignoring codec for track:', track);
												} else if (typeof track.codec !== 'string') {
													Janus.warn('Invalid codec value, ignoring for track:', track);
												} else {
													(function () {
														var mimeType = kind + '/' + track.codec.toLowerCase();
														var codecs = RTCRtpReceiver.getCapabilities(kind).codecs.filter(function (codec) {
															return codec.mimeType.toLowerCase() === mimeType;
														});
														if (!codecs || codecs.length === 0) {
															Janus.warn('Codec not supported in this browser for this track, ignoring:', track);
														} else {
															try {
																transceiver.setCodecPreferences(codecs);
															} catch (err) {
																Janus.warn('Failed enforcing codec for this ' + kind + ' track:', err);
															}
														}
													})();
												}
											}
											// Check if insertable streams are involved
											if (transceiver.receiver && track.transforms && track.transforms.receiver) {
												receiverStreams = null;

												if (RTCRtpReceiver.prototype.createEncodedStreams) {
													receiverStreams = transceiver.receiver.createEncodedStreams();
												} else if (RTCRtpReceiver.prototype.createAudioEncodedStreams || RTCRtpReceiver.prototype.createEncodedVideoStreams) {
													if (kind === 'audio') {
														receiverStreams = transceiver.receiver.createEncodedAudioStreams();
													} else if (kind === 'video') {
														receiverStreams = transceiver.receiver.createEncodedVideoStreams();
													}
												}
												if (receiverStreams) {
													console.log('Insertable Streams receiver transform:', receiverStreams);
													if (receiverStreams.readableStream && receiverStreams.writableStream) {
														receiverStreams.readableStream.pipeThrough(track.transforms.receiver).pipeTo(receiverStreams.writableStream);
													} else if (receiverStreams.readable && receiverStreams.writable) {
														receiverStreams.readable.pipeThrough(track.transforms.receiver).pipeTo(receiverStreams.writable);
													}
												}
											}
										}
									}

								case 122:
									if (!(trackId && config.myStream)) {
										context$3$0.next = 167;
										break;
									}

									rt = null;

									if (!(kind === 'audio' && config.myStream.getAudioTracks() && config.myStream.getAudioTracks().length)) {
										context$3$0.next = 146;
										break;
									}

									_iteratorNormalCompletion11 = true;
									_didIteratorError11 = false;
									_iteratorError11 = undefined;
									context$3$0.prev = 128;

									for (_iterator11 = config.myStream.getAudioTracks()[Symbol.iterator](); !(_iteratorNormalCompletion11 = (_step11 = _iterator11.next()).done); _iteratorNormalCompletion11 = true) {
										t = _step11.value;

										if (t.id === trackId) {
											rt = t;
											Janus.log('Removing audio track:', rt);
										}
									}
									context$3$0.next = 136;
									break;

								case 132:
									context$3$0.prev = 132;
									context$3$0.t0 = context$3$0["catch"](128);
									_didIteratorError11 = true;
									_iteratorError11 = context$3$0.t0;

								case 136:
									context$3$0.prev = 136;
									context$3$0.prev = 137;

									if (!_iteratorNormalCompletion11 && _iterator11["return"]) {
										_iterator11["return"]();
									}

								case 139:
									context$3$0.prev = 139;

									if (!_didIteratorError11) {
										context$3$0.next = 142;
										break;
									}

									throw _iteratorError11;

								case 142:
									return context$3$0.finish(139);

								case 143:
									return context$3$0.finish(136);

								case 144:
									context$3$0.next = 166;
									break;

								case 146:
									if (!(kind === 'video' && config.myStream.getVideoTracks() && config.myStream.getVideoTracks().length)) {
										context$3$0.next = 166;
										break;
									}

									_iteratorNormalCompletion12 = true;
									_didIteratorError12 = false;
									_iteratorError12 = undefined;
									context$3$0.prev = 150;

									for (_iterator12 = config.myStream.getVideoTracks()[Symbol.iterator](); !(_iteratorNormalCompletion12 = (_step12 = _iterator12.next()).done); _iteratorNormalCompletion12 = true) {
										t = _step12.value;

										if (t.id === trackId) {
											rt = t;
											Janus.log('Removing video track:', rt);
										}
									}
									context$3$0.next = 158;
									break;

								case 154:
									context$3$0.prev = 154;
									context$3$0.t1 = context$3$0["catch"](150);
									_didIteratorError12 = true;
									_iteratorError12 = context$3$0.t1;

								case 158:
									context$3$0.prev = 158;
									context$3$0.prev = 159;

									if (!_iteratorNormalCompletion12 && _iterator12["return"]) {
										_iterator12["return"]();
									}

								case 161:
									context$3$0.prev = 161;

									if (!_didIteratorError12) {
										context$3$0.next = 164;
										break;
									}

									throw _iteratorError12;

								case 164:
									return context$3$0.finish(161);

								case 165:
									return context$3$0.finish(158);

								case 166:
									if (rt) {
										// Remove the track and notify the application
										try {
											config.myStream.removeTrack(rt);
											pluginHandle.onlocaltrack(rt, false);
										} catch (e) {
											Janus.error("Error calling onlocaltrack on removal for renegotiation", e);
										}
										// Close the old track (unless we've been asked not to)
										if (rt.dontStop !== true) {
											try {
												rt.stop();
											} catch (e) {}
										}
									}

								case 167:
									if (nt) {
										// FIXME Add the new track locally
										config.myStream.addTrack(nt);
										// Notify the application about the new local track, if any
										nt.onended = function (ev) {
											Janus.log('Local track removed:', ev);
											try {
												pluginHandle.onlocaltrack(ev.target, false);
											} catch (e) {
												Janus.error("Error calling onlocaltrack following end", e);
											}
										};
										try {
											pluginHandle.onlocaltrack(nt, true);
										} catch (e) {
											Janus.error("Error calling onlocaltrack for track add", e);
										}
									}
									// Update the direction of the transceiver
									if (transceiver) {
										curdir = transceiver.direction, newdir = null;
										send = nt && transceiver.sender.track, recv = track.recv !== false && transceiver.receiver.track;

										if (send && recv) newdir = 'sendrecv';else if (send && !recv) newdir = 'sendonly';else if (!send && recv) newdir = 'recvonly';else if (!send && !recv) newdir = 'inactive';
										if (newdir && newdir !== curdir) {
											Janus.warn('Changing direction of transceiver to ' + newdir + ' (was ' + curdir + ')', track);
											if (transceiver.setDirection) transceiver.setDirection(newdir);else transceiver.direction = newdir;
										}
									}

								case 169:
								case "end":
									return context$3$0.stop();
							}
						}, null, _this, [[128, 132, 136, 144], [137,, 139, 143], [150, 154, 158, 166], [159,, 161, 165]]);
					};

					_iterator10 = tracks[Symbol.iterator]();

				case 75:
					if (_iteratorNormalCompletion10 = (_step10 = _iterator10.next()).done) {
						context$2$0.next = 84;
						break;
					}

					context$2$0.next = 78;
					return regeneratorRuntime.awrap(_loop());

				case 78:
					_ret4 = context$2$0.sent;

					if (!(_ret4 === "continue")) {
						context$2$0.next = 81;
						break;
					}

					return context$2$0.abrupt("continue", 81);

				case 81:
					_iteratorNormalCompletion10 = true;
					context$2$0.next = 75;
					break;

				case 84:
					context$2$0.next = 90;
					break;

				case 86:
					context$2$0.prev = 86;
					context$2$0.t2 = context$2$0["catch"](72);
					_didIteratorError10 = true;
					_iteratorError10 = context$2$0.t2;

				case 90:
					context$2$0.prev = 90;
					context$2$0.prev = 91;

					if (!_iteratorNormalCompletion10 && _iterator10["return"]) {
						_iterator10["return"]();
					}

				case 93:
					context$2$0.prev = 93;

					if (!_didIteratorError10) {
						context$2$0.next = 96;
						break;
					}

					throw _iteratorError10;

				case 96:
					return context$2$0.finish(93);

				case 97:
					return context$2$0.finish(90);

				case 98:
					if (openedConsentDialog) pluginHandle.consentDialog(false);

				case 99:
				case "end":
					return context$2$0.stop();
			}
		}, null, this, [[16, 36, 40, 48], [41,, 43, 47], [52, 56, 60, 68], [61,, 63, 67], [72, 86, 90, 98], [91,, 93, 97]]);
	}

	function _getLocalTracks(handleId) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn('Invalid handle');
			return null;
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.pc) {
			Janus.warn('Invalid PeerConnection');
			return null;
		}
		var tracks = [];
		var transceivers = config.pc.getTransceivers();
		var _iteratorNormalCompletion13 = true;
		var _didIteratorError13 = false;
		var _iteratorError13 = undefined;

		try {
			for (var _iterator13 = transceivers[Symbol.iterator](), _step13; !(_iteratorNormalCompletion13 = (_step13 = _iterator13.next()).done); _iteratorNormalCompletion13 = true) {
				var tr = _step13.value;

				var track = null;
				if (tr.sender && tr.sender.track) {
					track = { mid: tr.mid };
					track.type = tr.sender.track.kind;
					track.id = tr.sender.track.id;
					track.label = tr.sender.track.label;
				}
				if (track) tracks.push(track);
			}
		} catch (err) {
			_didIteratorError13 = true;
			_iteratorError13 = err;
		} finally {
			try {
				if (!_iteratorNormalCompletion13 && _iterator13["return"]) {
					_iterator13["return"]();
				}
			} finally {
				if (_didIteratorError13) {
					throw _iteratorError13;
				}
			}
		}

		return tracks;
	}

	function _getRemoteTracks(handleId) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn('Invalid handle');
			return null;
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.pc) {
			Janus.warn('Invalid PeerConnection');
			return null;
		}
		var tracks = [];
		var transceivers = config.pc.getTransceivers();
		var _iteratorNormalCompletion14 = true;
		var _didIteratorError14 = false;
		var _iteratorError14 = undefined;

		try {
			for (var _iterator14 = transceivers[Symbol.iterator](), _step14; !(_iteratorNormalCompletion14 = (_step14 = _iterator14.next()).done); _iteratorNormalCompletion14 = true) {
				var tr = _step14.value;

				var track = null;
				if (tr.receiver && tr.receiver.track) {
					track = { mid: tr.mid };
					track.type = tr.receiver.track.kind;
					track.id = tr.receiver.track.id;
					track.label = tr.receiver.track.label;
				}
				if (track) tracks.push(track);
			}
		} catch (err) {
			_didIteratorError14 = true;
			_iteratorError14 = err;
		} finally {
			try {
				if (!_iteratorNormalCompletion14 && _iterator14["return"]) {
					_iterator14["return"]();
				}
			} finally {
				if (_didIteratorError14) {
					throw _iteratorError14;
				}
			}
		}

		return tracks;
	}

	function _getVolume(handleId, mid, remote, result) {
		result = typeof result == "function" ? result : Janus.noop;
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			result(0);
			return;
		}
		var stream = remote ? "remote" : "local";
		var config = pluginHandle.webrtcStuff;
		if (!config.volume[stream]) config.volume[stream] = { value: 0 };
		// Start getting the volume, if audioLevel in getStats is supported (apparently
		// they're only available in Chrome/Safari right now: https://webrtc-stats.callstats.io/)
		if (config.pc && config.pc.getStats && (Janus.webRTCAdapter.browserDetails.browser === "chrome" || Janus.webRTCAdapter.browserDetails.browser === "safari")) {
			// Are we interested in a mid in particular?
			var query = config.pc;
			if (mid) {
				var transceiver = config.pc.getTransceivers().find(function (t) {
					return t.mid === mid && t.receiver.track.kind === "audio";
				});
				if (!transceiver) {
					Janus.warn("No audio transceiver with mid " + mid);
					result(0);
					return;
				}
				if (remote && !transceiver.receiver) {
					Janus.warn("Remote transceiver track unavailable");
					result(0);
					return;
				} else if (!remote && !transceiver.sender) {
					Janus.warn("Local transceiver track unavailable");
					result(0);
					return;
				}
				query = remote ? transceiver.receiver : transceiver.sender;
			}
			query.getStats().then(function (stats) {
				stats.forEach(function (res) {
					if (!res || res.kind !== "audio") return;
					if (remote && !res.remoteSource || !remote && res.type !== "media-source") return;
					result(res.audioLevel ? res.audioLevel : 0);
				});
			});
			return config.volume[stream].value;
		} else {
			// audioInputLevel and audioOutputLevel seem only available in Chrome? audioLevel
			// seems to be available on Chrome and Firefox, but they don't seem to work
			Janus.warn("Getting the " + stream + " volume unsupported by browser");
			result(0);
			return;
		}
	}

	function isMuted(handleId, mid, video) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			return true;
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.pc) {
			Janus.warn("Invalid PeerConnection");
			return true;
		}
		if (!config.myStream) {
			Janus.warn("Invalid local MediaStream");
			return true;
		}
		if (video) {
			// Check video track
			if (!config.myStream.getVideoTracks() || config.myStream.getVideoTracks().length === 0) {
				Janus.warn("No video track");
				return true;
			}
			if (mid) {
				var transceiver = config.pc.getTransceivers().find(function (t) {
					return t.mid === mid && t.receiver.track.kind === "video";
				});
				if (!transceiver) {
					Janus.warn("No video transceiver with mid " + mid);
					return true;
				}
				if (!transceiver.sender || !transceiver.sender.track) {
					Janus.warn("No video sender with mid " + mid);
					return true;
				}
				return !transceiver.sender.track.enabled;
			} else {
				return !config.myStream.getVideoTracks()[0].enabled;
			}
		} else {
			// Check audio track
			if (!config.myStream.getAudioTracks() || config.myStream.getAudioTracks().length === 0) {
				Janus.warn("No audio track");
				return true;
			}
			if (mid) {
				var transceiver = config.pc.getTransceivers().find(function (t) {
					return t.mid === mid && t.receiver.track.kind === "audio";
				});
				if (!transceiver) {
					Janus.warn("No audio transceiver with mid " + mid);
					return true;
				}
				if (!transceiver.sender || !transceiver.sender.track) {
					Janus.warn("No audio sender with mid " + mid);
					return true;
				}
				return !transceiver.sender.track.enabled;
			} else {
				return !config.myStream.getAudioTracks()[0].enabled;
			}
		}
	}

	function mute(handleId, mid, video, mute) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			return false;
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.pc) {
			Janus.warn("Invalid PeerConnection");
			return false;
		}
		if (!config.myStream) {
			Janus.warn("Invalid local MediaStream");
			return false;
		}
		if (video) {
			// Mute/unmute video track
			if (!config.myStream.getVideoTracks() || config.myStream.getVideoTracks().length === 0) {
				Janus.warn("No video track");
				return false;
			}
			if (mid) {
				var transceiver = config.pc.getTransceivers().find(function (t) {
					return t.mid === mid && t.receiver.track.kind === "video";
				});
				if (!transceiver) {
					Janus.warn("No video transceiver with mid " + mid);
					return false;
				}
				if (!transceiver.sender || !transceiver.sender.track) {
					Janus.warn("No video sender with mid " + mid);
					return false;
				}
				transceiver.sender.track.enabled = mute ? false : true;
			} else {
				var _iteratorNormalCompletion15 = true;
				var _didIteratorError15 = false;
				var _iteratorError15 = undefined;

				try {
					for (var _iterator15 = config.myStream.getVideoTracks()[Symbol.iterator](), _step15; !(_iteratorNormalCompletion15 = (_step15 = _iterator15.next()).done); _iteratorNormalCompletion15 = true) {
						var videostream = _step15.value;

						videostream.enabled = !mute;
					}
				} catch (err) {
					_didIteratorError15 = true;
					_iteratorError15 = err;
				} finally {
					try {
						if (!_iteratorNormalCompletion15 && _iterator15["return"]) {
							_iterator15["return"]();
						}
					} finally {
						if (_didIteratorError15) {
							throw _iteratorError15;
						}
					}
				}
			}
		} else {
			// Mute/unmute audio track
			if (!config.myStream.getAudioTracks() || config.myStream.getAudioTracks().length === 0) {
				Janus.warn("No audio track");
				return false;
			}
			if (mid) {
				var transceiver = config.pc.getTransceivers().find(function (t) {
					return t.mid === mid && t.receiver.track.kind === "audio";
				});
				if (!transceiver) {
					Janus.warn("No audio transceiver with mid " + mid);
					return false;
				}
				if (!transceiver.sender || !transceiver.sender.track) {
					Janus.warn("No audio sender with mid " + mid);
					return false;
				}
				transceiver.sender.track.enabled = mute ? false : true;
			} else {
				var _iteratorNormalCompletion16 = true;
				var _didIteratorError16 = false;
				var _iteratorError16 = undefined;

				try {
					for (var _iterator16 = config.myStream.getAudioTracks()[Symbol.iterator](), _step16; !(_iteratorNormalCompletion16 = (_step16 = _iterator16.next()).done); _iteratorNormalCompletion16 = true) {
						var audiostream = _step16.value;

						audiostream.enabled = !mute;
					}
				} catch (err) {
					_didIteratorError16 = true;
					_iteratorError16 = err;
				} finally {
					try {
						if (!_iteratorNormalCompletion16 && _iterator16["return"]) {
							_iterator16["return"]();
						}
					} finally {
						if (_didIteratorError16) {
							throw _iteratorError16;
						}
					}
				}
			}
		}
		return true;
	}

	function _getBitrate(handleId, mid) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn("Invalid handle");
			return "Invalid handle";
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.pc) return "Invalid PeerConnection";
		// Start getting the bitrate, if getStats is supported
		if (config.pc.getStats) {
			var _ret7 = (function () {
				var query = config.pc;
				var target = mid ? mid : "default";
				if (mid) {
					var transceiver = config.pc.getTransceivers().find(function (t) {
						return t.mid === mid && t.receiver.track.kind === "video";
					});
					if (!transceiver) {
						Janus.warn("No video transceiver with mid " + mid);
						return {
							v: "No video transceiver with mid " + mid
						};
					}
					if (!transceiver.receiver) {
						Janus.warn("No video receiver with mid " + mid);
						return {
							v: "No video receiver with mid " + mid
						};
					}
					query = transceiver.receiver;
				}
				if (!config.bitrate[target]) {
					config.bitrate[target] = {
						timer: null,
						bsnow: null,
						bsbefore: null,
						tsnow: null,
						tsbefore: null,
						value: "0 kbits/sec"
					};
				}
				if (!config.bitrate[target].timer) {
					Janus.log("Starting bitrate timer" + (mid ? " for mid " + mid : "") + " (via getStats)");
					config.bitrate[target].timer = setInterval(function () {
						query.getStats().then(function (stats) {
							stats.forEach(function (res) {
								if (!res) return;
								var inStats = false;
								// Check if these are statistics on incoming media
								if ((res.mediaType === "video" || res.id.toLowerCase().indexOf("video") > -1) && res.type === "inbound-rtp" && res.id.indexOf("rtcp") < 0) {
									// New stats
									inStats = true;
								} else if (res.type == 'ssrc' && res.bytesReceived && (res.googCodecName === "VP8" || res.googCodecName === "")) {
									// Older Chromer versions
									inStats = true;
								}
								// Parse stats now
								if (inStats) {
									config.bitrate[target].bsnow = res.bytesReceived;
									config.bitrate[target].tsnow = res.timestamp;
									if (config.bitrate[target].bsbefore === null || config.bitrate[target].tsbefore === null) {
										// Skip this round
										config.bitrate[target].bsbefore = config.bitrate[target].bsnow;
										config.bitrate[target].tsbefore = config.bitrate[target].tsnow;
									} else {
										// Calculate bitrate
										var timePassed = config.bitrate[target].tsnow - config.bitrate[target].tsbefore;
										if (Janus.webRTCAdapter.browserDetails.browser === "safari") timePassed = timePassed / 1000; // Apparently the timestamp is in microseconds, in Safari
										var bitRate = Math.round((config.bitrate[target].bsnow - config.bitrate[target].bsbefore) * 8 / timePassed);
										if (Janus.webRTCAdapter.browserDetails.browser === "safari") bitRate = parseInt(bitRate / 1000);
										config.bitrate[target].value = bitRate + ' kbits/sec';
										//~ Janus.log("Estimated bitrate is " + config.bitrate.value);
										config.bitrate[target].bsbefore = config.bitrate[target].bsnow;
										config.bitrate[target].tsbefore = config.bitrate[target].tsnow;
									}
								}
							});
						});
					}, 1000);
					return {
						v: "0 kbits/sec"
					}; // We don't have a bitrate value yet
				}
				return {
					v: config.bitrate[target].value
				};
			})();

			if (typeof _ret7 === "object") return _ret7.v;
		} else {
			Janus.warn("Getting the video bitrate unsupported by browser");
			return "Feature unsupported by browser";
		}
	}

	function setBitrate(handleId, mid, bitrate) {
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle || !pluginHandle.webrtcStuff) {
			Janus.warn('Invalid handle');
			return;
		}
		var config = pluginHandle.webrtcStuff;
		if (!config.pc) {
			Janus.warn('Invalid PeerConnection');
			return;
		}
		var transceiver = config.pc.getTransceivers().find(function (t) {
			return t.mid === mid;
		});
		if (!transceiver) {
			Janus.warn('No transceiver with mid', mid);
			return;
		}
		if (!transceiver.sender) {
			Janus.warn('No sender for transceiver with mid', mid);
			return;
		}
		var params = transceiver.sender.getParameters();
		if (!params || !params.encodings || params.encodings.length === 0) {
			Janus.warn('No parameters encodings');
		} else if (params.encodings.length > 1) {
			Janus.warn('Ignoring bitrate for simulcast track, use sendEncodings for that');
		} else if (isNaN(bitrate) || bitrate < 0) {
			Janus.warn('Invalid bitrate (must be a positive integer)');
		} else {
			params.encodings[0].maxBitrate = bitrate;
			transceiver.sender.setParameters(params);
		}
	}

	function webrtcError(error) {
		Janus.error("WebRTC error:", error);
	}

	function cleanupWebrtc(handleId, hangupRequest) {
		Janus.log("Cleaning WebRTC stuff");
		var pluginHandle = pluginHandles[handleId];
		if (!pluginHandle) {
			// Nothing to clean
			return;
		}
		var config = pluginHandle.webrtcStuff;
		if (config) {
			if (hangupRequest === true) {
				// Send a hangup request (we don't really care about the response)
				var request = { "janus": "hangup", "transaction": Janus.randomString(12) };
				if (pluginHandle.token) request["token"] = pluginHandle.token;
				if (apisecret) request["apisecret"] = apisecret;
				Janus.debug("Sending hangup request (handle=" + handleId + "):");
				Janus.debug(request);
				if (websockets) {
					request["session_id"] = sessionId;
					request["handle_id"] = handleId;
					ws.send(JSON.stringify(request));
				} else {
					Janus.httpAPICall(server + "/" + sessionId + "/" + handleId, {
						verb: 'POST',
						withCredentials: withCredentials,
						body: request
					});
				}
			}
			// Cleanup stack
			if (config.volume) {
				if (config.volume["local"] && config.volume["local"].timer) clearInterval(config.volume["local"].timer);
				if (config.volume["remote"] && config.volume["remote"].timer) clearInterval(config.volume["remote"].timer);
			}
			for (var i in config.bitrate) {
				if (config.bitrate[i].timer) clearInterval(config.bitrate[i].timer);
			}
			config.bitrate = {};
			if (!config.streamExternal && config.myStream) {
				Janus.log("Stopping local stream tracks");
				Janus.stopAllTracks(config.myStream);
			}
			config.streamExternal = false;
			config.myStream = null;
			// Close PeerConnection
			try {
				config.pc.close();
			} catch (e) {
				// Do nothing
			}
			config.pc = null;
			config.candidates = null;
			config.mySdp = null;
			config.remoteSdp = null;
			config.iceDone = false;
			config.dataChannel = {};
			config.dtmfSender = null;
			config.insertableStreams = false;
		}
		pluginHandle.oncleanup();
	}

	function isTrickleEnabled(trickle) {
		Janus.debug("isTrickleEnabled:", trickle);
		return trickle === false ? false : true;
	}
}

exports["default"] = Janus;
module.exports = exports["default"];

// If the deprecated media was provided instead of tracks, translate it

// Find the video track and add simulcast/SVC info there

// Check that callbacks.array is a valid array

// Get the plugin handle

// Capture devices and setup tracks, if needed

// Create offer or answer now (depending on the context)

// Create the answer now

// Nothing to do

// Check if we can/should group getUserMedia calls

// Easy enough: create a datachannel if we don't have one already

// Capture the new track, if we need to

// We did a getUserMedia before already

// Use getDisplayMedia

// Replace the track

// Update the transceiver direction

// Override maximum bitrate

// Override maximum framerate

// There's a sender transform, set it on the transceiver sender

// There's a receiver transform, set it on the transceiver receiver

// There's a receiver transform, set it on the transceiver receiver
// Get rid of the old track
// FIXME We should probably do this *before* capturing the new
// track, since this prevents, for instance, just changing the
// resolution of the same webcam we're capturing already (the
// existing resolution would be returned, or an overconstrained
// error). On the other end, closing the track before we capture
// the new device means we'd end up with a period of time where
// no video is sent (changing device takes some time), and
// media would be stopped entirely in case capturing the new
// device results in an error. To keep things simpler, we're
// doing it after: we can make this configurable in the future.